# author: zimri.leisher
# created on: Jan 27, 2025

# allow us to use bracketed types
from __future__ import annotations
import json as js
from pathlib import Path
from argparse import ArgumentParser
from typing import Any
from fprime_gds.common.loaders.prm_json_loader import PrmJsonLoader
from fprime_gds.common.templates.prm_template import PrmTemplate
from fprime.common.models.serialize.type_base import BaseType
from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.bool_type import BoolType
from fprime.common.models.serialize.enum_type import EnumType
from fprime.common.models.serialize.numerical_types import (
    F32Type,
    F64Type,
    I8Type,
    I16Type,
    I32Type,
    I64Type,
    U8Type,
    U16Type,
    U32Type,
    U64Type,
)
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.string_type import StringType

FW_PRM_ID_TYPE_SIZE = 4 # serialized size of the FwPrmIdType


def instantiate_prm_type(prm_val_json, prm_type: type[BaseType]):
    """given a parameter type and its value in json form, instantiate the type
    with the value, or raise an exception if the json is not compatible"""
    prm_instance = prm_type()
    if isinstance(prm_instance, BoolType):
        value = str(prm_val_json).lower().strip()
        if value in {"true", "yes"}:
            av = True
        elif value in {"false", "no"}:
            av = False
        else:
            raise RuntimeError("Param value is not a valid boolean")
        prm_instance.val = av
    elif isinstance(prm_instance, EnumType):
        prm_instance.val = prm_val_json
    elif isinstance(prm_instance, (F64Type, F32Type)):
        prm_instance.val = float(prm_val_json)
    elif isinstance(
        prm_instance,
        (I64Type, U64Type, I32Type, U32Type, I16Type, U16Type, I8Type, U8Type),
    ):
        prm_instance.val = int(prm_val_json, 0) if isinstance(prm_val_json, str) else int(prm_val_json)
    elif isinstance(prm_instance, StringType):
        prm_instance.val = prm_val_json
    elif isinstance(prm_instance, (ArrayType, SerializableType)):
        prm_instance.val = prm_val_json
    else:
        raise RuntimeError(
            "Param value could not be converted to type object"
        )
    return prm_instance


def parsed_json_to_dat(templates_and_values: list[tuple[PrmTemplate, Any]]) -> bytes:
    """convert a list of (PrmTemplate, prm value json) to serialized bytes for a PrmDb"""
    serialized = bytes()
    for template_and_value in templates_and_values:
        template, json_value = template_and_value
        prm_instance = instantiate_prm_type(json_value, template.prm_type_obj)

        prm_instance_bytes = prm_instance.serialize()

        # see https://github.com/nasa/fprime/blob/devel/Svc/PrmDb/docs/sdd.md#32-functional-description
        # for an explanation of the binary format of parameters in the .dat file

        # delimiter
        serialized += b"\xA5"

        record_size = FW_PRM_ID_TYPE_SIZE + len(prm_instance_bytes)

        # size of following data
        serialized += record_size.to_bytes(length=4, byteorder="big")
        # id of param
        serialized += template.prm_id.to_bytes(length=4, byteorder="big")
        # value of param
        serialized += prm_instance_bytes
    return serialized


def parsed_json_to_seq(templates_and_values: list[tuple[PrmTemplate, dict]], include_save=False) -> list[str]:
    """convert a list of (PrmTemplate, prm value json) to a command sequence for the CmdSequencer.
    Returns a list of lines in the sequence."""
    cmds = []
    cmds.append("; Autocoded sequence file from JSON")
    for template_and_value in templates_and_values:
        template, json_value = template_and_value
        set_cmd_name = template.comp_name + "." + template.prm_name.upper() + "_PRM_SET"
        cmd = "R00:00:00 " + set_cmd_name + " " + str(json_value)
        cmds.append(cmd)
        if include_save:
            save_cmd = template.comp_name + "." + template.prm_name.upper() + "_PRM_SAVE"
            cmds.append(save_cmd)
    return cmds



def parse_json(param_value_json, name_dict: dict[str, PrmTemplate], include_implicit_defaults=False) -> list[tuple[PrmTemplate, dict]]:
    """
    param_value_json: the json object read from the .json file
    name_dict: a dictionary of (fqn param name, PrmTemplate) pairs
    include_implicit_defaults: whether or not to also include default values from the name dict
                               if no value was specified in the json
    @return a list of tuple of param template and the intended param value (in form of json dict)
    """
    # first, check the json for errors
    for component_name in param_value_json:
        for param_name in param_value_json[component_name]:
            fqn_param_name = component_name + "." + param_name
            param_temp: PrmTemplate = name_dict.get(fqn_param_name, None)
            if not param_temp:
                raise RuntimeError(
                    "Unable to find param "
                    + fqn_param_name
                    + " in dictionary"
                )

    # okay, now iterate over the dict
    templates_to_values = []
    for fqn_param_name, prm_template in name_dict.items():

        prm_val = None

        if include_implicit_defaults:
            # there is a default value
            prm_val = prm_template.prm_default_val
        
        comp_json = param_value_json.get(prm_template.comp_name, None)
        if comp_json:
            # if there is an entry for the component
            if prm_template.prm_name in comp_json:
                # if there is an entry for this param
                # get the value
                prm_val = comp_json[prm_template.prm_name]
        
        if not prm_val:
            # not writing a val for this prm
            continue

        templates_to_values.append((prm_template, prm_val))

    return templates_to_values


def main():
    arg_parser = ArgumentParser()
    subparsers = arg_parser.add_subparsers(dest="subcmd", required=True)


    json_to_dat = subparsers.add_parser("dat", help="Compiles .json files into param DB .dat files")
    json_to_dat.add_argument(
        "json_file", type=Path, help="The .json file to turn into a .dat file", default=None
    )
    json_to_dat.add_argument(
        "--dictionary",
        "-d",
        type=Path,
        help="The dictionary file of the FSW",
        required=True,
    )
    json_to_dat.add_argument("--defaults", action="store_true", help="Whether or not to implicitly include default parameter values in the output")
    json_to_dat.add_argument("--output", "-o", type=Path, help="The output file", default=None)


    json_to_seq = subparsers.add_parser("seq", help="Converts .json files into command sequence .seq files")
    json_to_seq.add_argument(
        "json_file", type=Path, help="The .json file to turn into a .seq file", default=None
    )
    json_to_seq.add_argument(
        "--dictionary",
        "-d",
        type=Path,
        help="The dictionary file of the FSW",
        required=True,
    )
    json_to_seq.add_argument("--defaults", action="store_true", help="Whether or not to implicitly include default parameter values in the output")
    json_to_seq.add_argument("--save", action="store_true", help="Whether or not to include the PRM_SAVE cmd in the output")
    json_to_seq.add_argument("--output", "-o", type=Path, help="The output file", default=None)


    args = arg_parser.parse_args()

    if args.json_file is None or not args.json_file.exists():
        print("Unable to find", args.json_file)
        exit(1)

    if args.json_file.is_dir():
        print("json-file is a dir", args.json_file)
        exit(1)

    if not args.dictionary.exists():
        print("Unable to find", args.dictionary)
        exit(1)

    output_format = args.subcmd

    # just compile the one file in place
    if args.output is None:
        output_path = args.json_file.with_suffix("." + output_format)
    else:
        output_path = args.output
    
    convert_json(args.json_file, args.dictionary, output_path, output_format, args.defaults, args.save)


def convert_json(json_file: Path, dictionary: Path, output: Path, output_format: str, implicit_defaults=False, include_save_cmd=False):

    print("Converting", json_file, "to", output, "(format: ." + output_format + ")")
    output.parent.mkdir(parents=True, exist_ok=True)

    json = js.loads(json_file.read_text())

    dict_parser = PrmJsonLoader(str(dictionary.resolve()))
    id_dict, name_dict, versions = dict_parser.construct_dicts(
        str(dictionary.resolve())
    )

    templates_to_values = parse_json(json, name_dict, implicit_defaults)

    if output_format == "dat":
        serialized_values = parsed_json_to_dat(templates_to_values)

        print("Done, writing to", output.resolve())
        output.write_bytes(serialized_values)
    elif output_format == "seq":
        sequence_cmds = parsed_json_to_seq(templates_to_values, include_save_cmd)
        print("Done, writing to", output.resolve())
        output.write_text("\n".join(sequence_cmds))
    else:
        raise RuntimeError("Invalid output format " + str(output_format))


if __name__ == "__main__":
    main()


"""
cmd_json_loader.py:

Loads flight dictionary (JSON) and returns id and mnemonic based Python dictionaries of commands

@author thomas-bc
"""

from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime_gds.common.loaders.json_loader import JsonLoader
from fprime_gds.common.data_types.exceptions import GdsDictionaryParsingException


class CmdJsonLoader(JsonLoader):
    """Class to load json based command dictionaries"""

    COMMANDS_FIELD = "commands"

    NAME = "name"
    OPCODE = "opcode"
    DESC = "annotation"
    PARAMETERS = "formalParams"

    def construct_dicts(self, _):
        """
         Constructs and returns python dictionaries keyed on id and name

        Args:
            _: Unused argument (inherited)
         Returns:
            A tuple with two command dictionaries (python type dict):
            (id_dict, name_dict). The keys are the events' id and name fields
            respectively and the values are CmdTemplate objects
        """
        id_dict = {}
        name_dict = {}

        if self.COMMANDS_FIELD not in self.json_dict:
            raise GdsDictionaryParsingException(
                f"Ground Dictionary missing '{self.COMMANDS_FIELD}' field: {str(self.json_file)}"
            )

        for cmd_dict in self.json_dict[self.COMMANDS_FIELD]:
            cmd_temp = self.construct_template_from_dict(cmd_dict)

            id_dict[cmd_temp.get_id()] = cmd_temp
            name_dict[cmd_temp.get_full_name()] = cmd_temp

        return (
            dict(sorted(id_dict.items())),
            dict(sorted(name_dict.items())),
            self.get_versions(),
        )

    def construct_template_from_dict(self, cmd_dict: dict) -> CmdTemplate:
        try:
            qualified_component_name, cmd_mnemonic = cmd_dict[self.NAME].rsplit('.', 1)
            cmd_opcode = cmd_dict[self.OPCODE]
            cmd_desc = cmd_dict.get(self.DESC)
        except ValueError as e:
            raise GdsDictionaryParsingException(
                f"Command dictionary entry malformed, expected name of the form '<QUAL_COMP_NAME>.<CMD_NAME>' in : {str(cmd_dict)}"
            )
        except KeyError as e:
            raise GdsDictionaryParsingException(
                f"{str(e)} key missing from Command dictionary entry: {str(cmd_dict)}"
            )
        # Parse Arguments
        cmd_args = []
        for param in cmd_dict.get(self.PARAMETERS, []):
            try:
                param_name = param["name"]
                param_type = self.parse_type(param["type"])
            except KeyError as e:
                raise GdsDictionaryParsingException(
                    f"{str(e)} key missing from Command parameter or its associated type in the dictionary: {str(param)}"
                )
            cmd_args.append(
                (
                    param_name,
                    param.get("annotation"),
                    param_type,
                )
            )
        return CmdTemplate(cmd_opcode, cmd_mnemonic, qualified_component_name, cmd_args, cmd_desc)



"""
comm.py:

This is the F prime communications adapter. This allows the F prime ground tool suite to interact with running F prime
deployments that exist on the other end of a "wire" (some communication bus). This is done with the following mechanics:

1. An adapter is instantiated to handle "read" and "write" functions against the wire
2. A framer/deframer is instantiated in order to frame/deframe those packets as transported across the wire.
3. "Uplink" and "Downlink" threads are created to loop on data from flight (F prime) and ground (F prime ground)
   interfaces ensuring that ground data is framed and written to the wire, and flight data is deframed and sent to the
   ground side.

Note: assuming the module containing the ground adapter has been imported, then this code should provide it as a CLI
      argument, removing the need to rewrite most of this class to use something different.

@author lestarch
"""

import logging
import signal
import sys
from pathlib import Path

# Required adapters built on standard tools
import fprime_gds.common.communication.adapters.base
import fprime_gds.common.communication.adapters.ip
import fprime_gds.common.communication.ground
import fprime_gds.common.logger
import fprime_gds.executables.cli
from fprime_gds.common.communication.updown import Downlinker, Uplinker
from fprime_gds.common.zmq_transport import ZmqGround
from fprime_gds.plugin.system import Plugins

# Uses non-standard PIP package pyserial, so test the waters before getting a hard-import crash
try:
    import fprime_gds.common.communication.adapters.uart
except ImportError:
    pass


LOGGER = logging.getLogger("comm")


def main():
    """
    Main program, degenerates into the run loop.

    :return: return code
    """
    # comm.py supports 2 and only 2 plugin categories
    Plugins.system(["communication", "framing"])
    args, _ = fprime_gds.executables.cli.ParserBase.parse_args(
        [
            fprime_gds.executables.cli.LogDeployParser,
            fprime_gds.executables.cli.MiddleWareParser,
            fprime_gds.executables.cli.CommParser,
            fprime_gds.executables.cli.PluginArgumentParser,
        ],
        description="F prime communications layer.",
        client=True,
    )
    if args.communication_selection == "none":
        print(
            "[ERROR] Comm adapter set to 'none'. Nothing to do but exit.",
            file=sys.stderr,
        )
        sys.exit(-1)

    # Create the handling components for either side of this script, adapter for hardware, and ground for the GDS side
    if args.zmq:
        ground = ZmqGround(args.zmq_transport)
    else:
        ground = fprime_gds.common.communication.ground.TCPGround(
            args.tts_addr, args.tts_port
        )

    adapter = Plugins.system().get_selected_class("communication")()

    # Set the framing class used and pass it to the uplink and downlink component constructions giving each a separate
    # instantiation
    framer_instance = Plugins.system().get_selected_class("framing")()
    LOGGER.info(
        "Starting uplinker/downlinker connecting to FSW using %s with %s",
        args.communication_selection,
        args.framing_selection,
    )
    discarded_file_handle = None
    try:
        if args.output_unframed_data == "-":
            discarded_file_handle = sys.stdout.buffer
        elif args.output_unframed_data is not None:
            discarded_file_handle_path = (
                Path(args.logs) / Path(args.output_unframed_data)
            ).resolve()
            try:
                discarded_file_handle = open(discarded_file_handle_path, "wb")
                LOGGER.info("Logging unframed data to %s", discarded_file_handle_path)
            except OSError:
                LOGGER.warning(
                    "Failed to open %s. Unframed data will be discarded.",
                    discarded_file_handle_path,
                )
        downlinker = Downlinker(
            adapter, ground, framer_instance, discarded=discarded_file_handle
        )
        uplinker = Uplinker(adapter, ground, framer_instance, downlinker)

        # Open resources for the handlers on either side, this prepares the resources needed for reading/writing data
        ground.open()
        adapter.open()

        # Finally start the processing of uplink and downlink
        downlinker.start()
        uplinker.start()
        LOGGER.debug("Uplinker and downlinker running")

        # Wait for shutdown event in the form of a KeyboardInterrupt then stop the processing, close resources,
        # and wait for everything to terminate as expected.
        def shutdown(*_):
            """Shutdown function for signals"""
            uplinker.stop()
            downlinker.stop()
            uplinker.join()
            downlinker.join()
            ground.close()
            adapter.close()

        signal.signal(signal.SIGTERM, shutdown)
        signal.signal(signal.SIGINT, shutdown)
        uplinker.join()
        downlinker.join()
    finally:
        if discarded_file_handle is not None and args.output_unframed_data != "-":
            discarded_file_handle.close()
    return 0


if __name__ == "__main__":
    sys.exit(main())



""" pytest_integration.py: F´ fixture API fixture

pytest uses fixtures to provide for extra functionality when writing tests. This fixture sets up the F´ stack allowing
our tests to use a configured test API without needing to create any specific setup to use that API. i.e. write a test
as follows:

```python
def test_my_test(fprime_test_api):
    ''' Perform my test '''
    fprime_test_api.send_and_assert_command(...)
```

Here a test (defined by starting the name with test_) uses the fprime_test_api fixture to perform the test.

@author lestarch
"""
import sys
from pathlib import Path
import pytest

from fprime_gds.common.testing_fw.api import IntegrationTestAPI
from fprime_gds.executables.cli import StandardPipelineParser

SEQUENCE_COUNTER = -1


def pytest_addoption(parser):
    """ Add fprime-gds options to the pytest parser

    Pytest allows users to add options to its parser. These options act very similar to argparse options and thus can be
    reused from the standard GDS command line processing. Note: pytest restricts the use of short flags (-[a-z]) thus we
    strip those from the standard cli processing. Long options must be supplied when testing using pytest.

    Args:
        parser: pytest style parser. Use "addoption" to add an option to it.
    """
    for flags, specifiers in StandardPipelineParser().get_arguments().items():
        # Reduce flags to only the long option (i.e. --something) form
        flags = [flag for flag in flags if flag.startswith("--")]
        parser.addoption(*flags, **specifiers)
        
    # Add an option to specify JUnit XML report file
    parser.addoption(
        "--junit-xml-file",
        action="store",
        default="report.xml",
        help="File to store JUnit XML report. [default: %(default)s]",
    )
    # Add an option to enable JUnit XML report generation to a specified location
    parser.addoption(
        "--gen-junitxml",
        action="store_true",
        help="Enable JUnitXML report generation to a specified location"
    )
    # Add an option to specify a json config file that maps components
    parser.addoption(
        "--deployment-config",
        action="store",
        help="Path to JSON configuration file for mapping deployment components"
    )

def pytest_configure(config):
    """ This is a hook to allow plugins and conftest files to perform initial configuration
    
    This hook is called for every initial conftest file after command line options have been parsed. After that, the 
    hook is called for other conftest files as they are registered.
    """
    # Create a JUnit XML report file to capture the test result in a specified location
    if config.getoption("--gen-junitxml"):
        config.option.xmlpath = Path(config.getoption("--logs")) / config.getoption("--junit-xml-file")

@pytest.fixture(scope='session')
def fprime_test_api_session(request):
    """ Create a session-level fprime test API

    This is a pytest session fixture. Using the options added above, this will parse the necessary options for
    connecting the standard pipeline to the running GDS. This pipeline is supplied to the fprime test API returned as
    the result of this fixture. This has several implications:
      1. APIs all use one connection to the GDS
      2. APIs and the connections are live across the whole pytest session. See fprime_test_api.

    Note: the implementation uses yield in order to setup the object and tear it down afterwards. This is as-recommended
    from pytest documentation.

    Args:
        request: standard pytest-supplied harness used for processing CLI arguments

    Return:
        fprime test API connected to the GDS.  Note: a second call will shut down that object.
    """
    pipeline_parser = StandardPipelineParser()
    pipeline = None
    api = None
    deployment_config = None
    try:
        # Parse the command line arguments into a client connection
        arg_ns = pipeline_parser.handle_arguments(request.config.known_args_namespace, client=True)

        # Build a new pipeline with the parsed and processed arguments
        pipeline = pipeline_parser.pipeline_factory(arg_ns, pipeline)

        # Get deployment configuration from command line arguments
        if request.config.option.deployment_config:
            deployment_config = request.config.option.deployment_config

        # Build and set up the integration test api
        api = IntegrationTestAPI(pipeline, deployment_config, arg_ns.logs)
        api.setup()

        # Return the API. Note: the second call here-in will begin after the yield and clean-up after the test
        yield api
    # In all cases, whether setup erred or proceeded to the yield, the teardown is now necessary
    finally:
        # Attempt to teardown the API to ensure we are clean after the fixture is created
        try:
            if api is not None:
                api.teardown()
        except Exception as exc:
            print(f"[WARNING] Exception in API teardown: {exc}", file=sys.stderr)
        # Attempt to shut down the pipeline connection
        try:
            if pipeline is not None:
                pipeline.disconnect()
        except Exception as exc:
            print(f"[WARNING] Exception in pipeline teardown: {exc}", file=sys.stderr)


@pytest.fixture(scope='function')
def fprime_test_api(fprime_test_api_session, request):
    """ Provide a per-testcase fixture

    Although the test API should exist across all testcases and thus be created at the "session" level, individual test
    cases need a clean test API that has logged the testcases has started. Thus, the session API is refined into a
    test case API by performing the test case setup consisting of:

    1. clearing current API history
    2. logging the test case name

    Args:
        fprime_test_api_session: session API to adjust for each test case
        request: pytest supplied request to get test name

    Return:
        test case specific session (identical to full session)
    """
    global SEQUENCE_COUNTER
    SEQUENCE_COUNTER += 1
    fprime_test_api_session.start_test_case(request.node.name, SEQUENCE_COUNTER)
    return fprime_test_api_session



####
# app.py:
#
# This file sets up the flask app, and registers the endpoints for the individual APIs supplied by
# this framework.
#
####
import logging
import os
import sys
import uuid

import flask

# Try to import Compress, but disable compression if not installed
try:
    from flask_compress import Compress
except ImportError:
    Compress = None

import fprime_gds.flask.channels

# Import the Flask API implementations
import fprime_gds.flask.commands
import fprime_gds.flask.errors
import fprime_gds.flask.events
import fprime_gds.flask.json
import fprime_gds.flask.logs
import fprime_gds.flask.sequence
import fprime_gds.flask.stats
import fprime_gds.flask.updown
from fprime_gds.executables.cli import ParserBase, StandardPipelineParser

from . import components

# Update logging to avoid redundant messages
logger = logging.getLogger("werkzeug")
logger.setLevel(logging.WARN)
logger = logging.getLogger("downlink")
logger.setLevel(logging.INFO)


def construct_app():
    """
    Constructs the Flask app by taking the following steps:

    1. Setup and configure the app
    2. Setup JSON encoding for Flask and flask_restful to handle F prime types natively
    3. Setup standard pipeline used throughout the system
    4. Create Restful API for registering flask items
    5. Register all restful endpoints

    :return: setup app
    """
    app = flask.Flask(__name__, static_url_path="")
    # Enable compression if it is installed
    if Compress is not None:
        compress = Compress()
        compress.init_app(app)

    app.config.from_object("fprime_gds.flask.default_settings")
    # Override defaults from python files specified in 'FP_FLASK_SETTINGS'
    if "FP_FLASK_SETTINGS" in os.environ:
        app.config.from_envvar("FP_FLASK_SETTINGS")

    # JSON encoding settings
    app.json.default = fprime_gds.flask.json.default
    app.config["RESTFUL_JSON"] = {"default": app.json.default}
    # Standard pipeline creation
    input_arguments = app.config["STANDARD_PIPELINE_ARGUMENTS"]
    args_ns, _ = ParserBase.parse_args(
        [StandardPipelineParser], "n/a", input_arguments, client=True
    )
    pipeline = components.setup_pipelined_components(app.debug, args_ns)

    # Restful API registration
    api = fprime_gds.flask.errors.setup_error_handling(app)

    # Application routes
    api.add_resource(
        fprime_gds.flask.commands.CommandDictionary,
        "/dictionary/commands",
        resource_class_args=[
            pipeline.dictionaries.command_name,
            pipeline.dictionaries.project_version,
            pipeline.dictionaries.framework_version,
            pipeline.dictionaries.metadata,
        ],
    )
    api.add_resource(
        fprime_gds.flask.commands.CommandHistory,
        "/commands",
        resource_class_args=[pipeline.histories.commands],
    )
    api.add_resource(
        fprime_gds.flask.commands.Command,
        "/commands/<command>",
        resource_class_args=[pipeline],
    )
    api.add_resource(
        fprime_gds.flask.events.EventDictionary,
        "/dictionary/events",
        resource_class_args=[
            pipeline.dictionaries.event_id,
            pipeline.dictionaries.project_version,
            pipeline.dictionaries.framework_version,
            pipeline.dictionaries.metadata,
        ],
    )
    api.add_resource(
        fprime_gds.flask.events.EventHistory,
        "/events",
        resource_class_args=[pipeline.histories.events],
    )
    api.add_resource(
        fprime_gds.flask.channels.ChannelDictionary,
        "/dictionary/channels",
        resource_class_args=[
            pipeline.dictionaries.channel_id,
            pipeline.dictionaries.project_version,
            pipeline.dictionaries.framework_version,
            pipeline.dictionaries.metadata,
        ],
    )
    api.add_resource(
        fprime_gds.flask.channels.ChannelHistory,
        "/channels",
        resource_class_args=[pipeline.histories.channels],
    )
    api.add_resource(
        fprime_gds.flask.updown.Destination,
        "/upload/destination",
        resource_class_args=[pipeline.files.uplinker],
    )
    api.add_resource(
        fprime_gds.flask.updown.FileUploads,
        "/upload/files",
        resource_class_args=[pipeline.files.uplinker, pipeline.up_store],
    )
    api.add_resource(
        fprime_gds.flask.updown.FileDownload,
        "/download/files",
        "/download/files/<string:source>",
        resource_class_args=[pipeline.files.downlinker],
    )
    api.add_resource(
        fprime_gds.flask.sequence.SequenceCompiler,
        "/sequence",
        resource_class_args=[
            args_ns.dictionary,
            pipeline.up_store,
            pipeline.files.uplinker,
            args_ns.remote_sequence_directory,
        ],
    )
    api.add_resource(
        fprime_gds.flask.stats.StatsBlob,
        "/stats",
        resource_class_args=[
            {
                "events": pipeline.histories.events,
                "channels": pipeline.histories.channels,
                "commands": pipeline.histories.commands,
            }
        ],
    )

    # Optionally serve log files
    if app.config["SERVE_LOGS"]:
        api.add_resource(
            fprime_gds.flask.logs.LogList,
            "/logdata",
            resource_class_args=[args_ns.logs],
        )
        api.add_resource(
            fprime_gds.flask.logs.LogFile,
            "/logdata/<name>",
            resource_class_args=[args_ns.logs],
        )
    return app, api


try:
    app, _ = construct_app()
except Exception as exc:
    print(f"[ERROR] {exc}", file=sys.stderr)
    sys.exit(1)


@app.errorhandler(Exception)
def handle_unexpected_error(error):
    status_code = 500
    response = {"errors": [fprime_gds.flask.errors.build_error_object(error)]}
    return flask.jsonify(response), status_code


@app.route("/js/<path:path>")
def files_serve(path):
    """
    A function used to serve the JS files needed for the GUI layers.

    :param path: path to the file (in terms of web browser)
    """
    return flask.send_from_directory("static/js", path)


@app.route("/")
def index():
    """
    A function used to serve the JS files needed for the GUI layers.
    """
    return flask.send_from_directory("static", "index.html")


@app.route("/logs")
def log():
    """
    A function used to serve the JS files needed for the GUI layers.
    """
    return flask.send_from_directory("static", "logs.html")


@app.route("/session")
def session():
    return flask.jsonify({"session": uuid.uuid4()}), 200


@app.after_request
def set_no_cache(response):
    """Set the no-cache header"""
    response.headers["Cache-Control"] = "no-cache"
    return response


# When running from the command line, this will allow the flask development server to launch
if __name__ == "__main__":
    app.run()



import unittest
from pathlib import Path

from fprime_gds.executables import utils


class TestFormatString(unittest.TestCase):

    def test_find_app_with_no_bin_dir_exits(self):
        path_with_no_bin = Path("")
        with self.assertRaises(SystemExit):
            utils.find_app(path_with_no_bin)

    def test_find_dict_with_no_dict_dir_exits(self):
        path_with_no_dict = Path("")
        with self.assertRaises(SystemExit):
            utils.find_app(path_with_no_dict)



"""
@brief Params Template class

Instances of this class describe a parameter of a component instance (not 
including a specific value)

@date Created January 27, 2025
@author Zimri Leisher

@bug Hopefully none
"""

from fprime.common.models.serialize.type_base import BaseType
from fprime.common.models.serialize.type_exceptions import TypeMismatchException

from . import data_template


class PrmTemplate(data_template.DataTemplate):
    """Class for param templates that describe parameters of component instances"""

    def __init__(
        self,
        prm_id: int,
        prm_name: str,
        comp_name: str,
        prm_type_obj: BaseType,
        prm_default_val,
    ):
        """
        Constructor

        Args:
            prm_id: the id of the parameter
            prm_name: the name of the parameter
            comp_name: the name of the component instance owning this parameter
            prm_type_obj: the instance of BaseType corresponding to the type of this parameter
            prm_default_val: the default value of this parameter, in raw JSON form
        """
        super().__init__()
        # Make sure correct types are passed
        if not isinstance(prm_id, int):
            raise TypeMismatchException(int, type(prm_id))

        if not isinstance(prm_name, str):
            raise TypeMismatchException(str, type(prm_name))

        if not isinstance(comp_name, str):
            raise TypeMismatchException(str, type(comp_name))

        if not issubclass(prm_type_obj, BaseType):
            raise TypeMismatchException(BaseType, prm_type_obj)

        # prm_default_val is an arbitrary type, likely a primitive or dict

        self.prm_id = prm_id
        self.prm_name = prm_name
        self.comp_name = comp_name
        self.prm_type_obj = prm_type_obj
        self.prm_default_val = prm_default_val

    def get_full_name(self):
        """
        Get the full name of this param

        Returns:
            The full name (component.param) for this param
        """
        return f"{self.comp_name}.{self.prm_name}"

    def get_id(self):
        return self.prm_id

    def get_name(self):
        return self.prm_name

    def get_comp_name(self):
        return self.comp_name

    def get_type_obj(self):
        return self.prm_type_obj



"""
@brief Loader class for importing xml based command dictionaries

@date Created July 23, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime_gds.common.data_types import exceptions
from fprime_gds.common.templates.cmd_template import CmdTemplate

# Custom Python Modules
from .xml_loader import XmlLoader


class CmdXmlLoader(XmlLoader):
    """Class to load xml based command dictionaries"""

    CMD_SECT = "commands"
    EVENT_SECT = "events"

    COMP_TAG = "component"
    MNEMONIC_TAG = "mnemonic"
    OPCODE_TAG = "opcode"
    DESC_TAG = "description"

    def construct_dicts(self, path):
        """
        Constructs and returns python dictionaries keyed on id and name

        This function should not be called directly, instead, use
        get_id_dict(path) and get_name_dict(path)

        Args:
            path: Path to the xml dictionary file containing command information

        Returns:
            A tuple with two command dictionaries (python type dict):
            (id_dict, name_dict). The keys are the events' id and name fields
            respectively and the values are CmdTemplate objects
        """
        xml_tree = self.get_xml_tree(path)
        versions = xml_tree.attrib.get("framework_version", "unknown"), xml_tree.attrib.get("project_version", "unknown")

        # Check if xml dict has commands section
        cmd_section = self.get_xml_section(self.CMD_SECT, xml_tree)
        if cmd_section is None:
            msg = f"Xml dict did not have a {self.EVENT_SECT} section"
            raise exceptions.GseControllerParsingException(
                msg
            )

        id_dict = {}
        name_dict = {}

        for cmd in cmd_section:
            cmd_dict = cmd.attrib

            cmd_comp = cmd_dict[self.COMP_TAG]
            cmd_mnemonic = cmd_dict[self.MNEMONIC_TAG]
            cmd_opcode = int(cmd_dict[self.OPCODE_TAG], base=16)

            cmd_desc = None
            if self.DESC_TAG in cmd_dict:
                cmd_desc = cmd_dict[self.DESC_TAG]

            # Parse Arguments
            args = self.get_args_list(cmd, xml_tree, f"{ cmd_comp }::{ cmd_mnemonic }")

            cmd_temp = CmdTemplate(cmd_opcode, cmd_mnemonic, cmd_comp, args, cmd_desc)

            id_dict[cmd_opcode] = cmd_temp
            name_dict[cmd_temp.get_full_name()] = cmd_temp

        return id_dict, name_dict, versions



"""
@brief Base class for all decoders. Defines the Decoder interface.

Decoders are responsible for taking in serialized data and parsing it into
objects. Decoders receive serialized data (that had a specific descriptor) from
a distributer that it has been registered to. The distributer will send the
binary data after removing any length and descriptor headers.

Example data that would be sent to a decoder that parses events or channels:
    +-------------------+---------------------+------------ - - -
    | ID (4 bytes) | Time Tag (11 bytes) | Data....
    +-------------------+---------------------+------------ - - -

This base class does not do any parsing, but instead acts as a pass through to
allow consumers to receive raw data.

@date Created June 29, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

import abc
import logging

import fprime_gds.common.handlers

LOGGER = logging.getLogger("decoder")


class DecodingException(Exception):
    """ Decoding Exception """


class Decoder(
    fprime_gds.common.handlers.DataHandler,
    fprime_gds.common.handlers.HandlerRegistrar,
    abc.ABC,
):
    """
    Base class for all decoder classes. This defines the "decode_api" function to allow for decoding of raw bytes. In
    addition it has a "data_callback" function implementation that decodes and sends out all results.
    """

    def data_callback(self, data, sender=None):
        """
        Data callback which calls the decode_api function exactly once. Then it passes the results to all registered
        consumer. This should only need to be overridden in extraordinary circumstances.

        :param data: data bytes to be decoded
        :param sender: (optional) sender id, otherwise None
        """
        try:
            decoded = self.decode_api(data)
        # Items already of DecodingException type should be reraised as-is
        except DecodingException:
            raise
        # Convert other exceptions into known DecodingException type
        except Exception as exc:
            raise DecodingException(str(exc))
        if decoded is not None:
            for item in decoded:
                self.send_to_all(item)
            return
        LOGGER.warning("Decoder of type %s produced 'None' decoded object", type(self))

    @abc.abstractmethod
    def decode_api(self, data):
        """
        Decodes the given data and returns the result.

        This function allows for non-registered code to call the same decoding
        code as is used to parse data passed to the data_callback function.

        :param data: binary data to decode
        :return: decoded data object
        """



""" fprime_encryption.framing.chain: implementation of a chained framer/deframer """
from abc import ABC, abstractmethod
from functools import reduce
from typing import Any, Dict, List, Type
from fprime_gds.common.communication.framing import FramerDeframer
from fprime_gds.common.communication.ccsds.space_data_link import SpaceDataLinkFramerDeframer
from fprime_gds.common.communication.ccsds.space_packet import SpacePacketFramerDeframer
from fprime_gds.plugin.definitions import gds_plugin



class ChainedFramerDeframer(FramerDeframer, ABC):
    """ Framer/deframer that is a composite of chained framer/deframers

    This Framer/Deframer will wrap a set of framer/deframers where the result of the frame and deframe options will pass
    from one to the other subsequently. The order is specified via the framing path and deframing will use the reverse
    order from specified.
    """
    def __init__(self, **kwargs):
        """ Initialize the chained framer/deframer from a framing-ordered set of children """
        frame_order_framer_deframers = [
            composite(**self.get_argument_subset(composite, kwargs))
            for composite in self.get_composites()
        ]
        self.framers = frame_order_framer_deframers[::1]
        self.deframers = frame_order_framer_deframers[::-1]

    @classmethod
    @abstractmethod
    def get_composites(cls) -> List[Type[FramerDeframer]]:
        """ Return a list of composites 
        Innermost FramerDeframer should be first in the list. """
        raise NotImplementedError(f"Subclasses of {cls.__name__} must implement get_composites")

    @staticmethod
    def get_argument_subset(composite: Type[FramerDeframer], argument_dictionary: Dict[str, Any]) -> Dict[str, Any]:
        """ Get an argument subset that is needed by composite

        For the composite, find the set of arguments that is needed by this composite and pull those out of the complete
        argument dictionary.

        Args:
            composite: class of a subtype of FramerDeframer
            argument_dictionary: dictionary of all input arguments
        """
        if not hasattr(composite, "get_arguments"):
            return {}
        needed_arguments = composite.get_arguments()
        needed_argument_destinations = [
            description["destination"] if "destination" in description else
                [dash_dash for dash_dash in flag if dash_dash.startswith("--")][0].lstrip("-").replace("-", "_")
            for flag, description in needed_arguments.items()
        ]
        return {name: argument_dictionary[name] for name in needed_argument_destinations}

    @classmethod
    def get_arguments(cls):
        """ Arguments to request from the CLI """
        all_arguments = {}
        for composite in cls.get_composites():
            all_arguments.update(composite.get_arguments() if hasattr(composite, "get_arguments") else {})
        return all_arguments

    @classmethod
    def check_arguments(cls, **kwargs):
        """ Check arguments from the CLI """
        for composite in cls.get_composites():
            subset_arguments = cls.get_argument_subset(composite, kwargs)
            if hasattr(composite, "check_arguments"):
                composite.check_arguments(**subset_arguments)

    def deframe(self, data, no_copy=False):
        """ Deframe via a chain of children deframers """
        packet = data[:] if not no_copy else data
        remaining = None
        discarded = b""

        for deframer in self.deframers:
            new_packet, new_remaining, new_discarded = deframer.deframe(packet, True)
            discarded += new_discarded
            remaining = new_remaining if remaining is None else remaining
            packet = new_packet
        return packet, remaining, discarded

    def frame(self, data):
        """ Frame via a chain of children framers """
        return reduce(lambda framed_data, framer: framer.frame(framed_data), self.framers, data)


@gds_plugin(FramerDeframer)
class SpacePacketSpaceDataLinkFramerDeframer(ChainedFramerDeframer):
    """ Space Data Link Protocol framing and deframing that has a data unit of Space Packets as the central """

    @classmethod
    def get_composites(cls) -> List[Type[FramerDeframer]]:
        """ Return the composite list of this chain 
        Innermost FramerDeframer should be first in the list. """
        return [
            SpacePacketFramerDeframer,
            SpaceDataLinkFramerDeframer
        ]

    @classmethod
    def get_name(cls):
        """ Name of this implementation provided to CLI """
        return "space-packet-space-data-link"



import unittest
import os

from fprime_gds.executables import data_product_writer

class TestRunDataProduct(unittest.TestCase):

    def test_data_product_parser(self):
        args = data_product_writer.parse_args(["test.bin", "dictionary.json"])
        assert args.binFile == "test.bin"
        assert args.jsonDict == "dictionary.json"

    def test_data_product_writer(self):
        binFiles = ["makeBool.bin",
                    "makeDataArray.bin",
                    "makeF32.bin",
                    "makeFppArray.bin",
                    "makeI32.bin",
                    "makeI8.bin",
                    "makeU32Array.bin",
                    "makeComplex.bin",
                    "makeEnum.bin",
                    "makeF64.bin",
                    "makeI16.bin",
                    "makeI64.bin",
                    "makeU32.bin",
                    "makeU8Array.bin"]
        
        # Specify the directory where the .bin files are located
        cwd_directory = os.getcwd()
        test_directory = os.path.dirname(os.path.abspath(__file__))

        # Loop through each file in the specified directory
        for filename in binFiles:
            print(f'Processing {filename}')
            bin_file = os.path.join(test_directory, "dp_writer_data", filename)
            dict_file = os.path.join(test_directory, "dp_writer_data", "dictionary.json")
            args = data_product_writer.parse_args([bin_file, dict_file])
            data_product_writer.DataProductWriter(args.jsonDict, args.binFile).process()
            #data_product_writer.process(args)

            # Check if the json file was created
            jsonFile = filename.replace('.bin', '.json')
            jsonFilePath = os.path.join(cwd_directory, jsonFile)
            self.assertTrue(os.path.exists(jsonFilePath))
    
            # If the json file exists, delete it
            if os.path.exists(jsonFilePath):
                    os.remove(jsonFile)


from __future__ import annotations
from abc import ABC
import inspect
from dataclasses import dataclass, field, fields
import traceback
from typing import Callable
import typing

from fprime_gds.common.fpy.bytecode.directives import (
    FLOAT_INEQUALITY_DIRECTIVES,
    MAX_SERIALIZABLE_REGISTER_SIZE,
    INT_SIGNED_INEQUALITY_DIRECTIVES,
    INT_UNSIGNED_INEQUALITY_DIRECTIVES,
    AndDirective,
    CmdDirective,
    DeserSerReg1Directive,
    DeserSerReg2Directive,
    DeserSerReg4Directive,
    DeserSerReg8Directive,
    Directive,
    FloatExtendDirective,
    IntEqualDirective,
    ExitDirective,
    FloatEqualDirective,
    FloatNotEqualDirective,
    GetPrmDirective,
    GetTlmDirective,
    GotoDirective,
    IfDirective,
    NotDirective,
    IntNotEqualDirective,
    OrDirective,
    SetSerRegDirective,
    SetRegDirective,
    SignedIntToFloatDirective,
    UnsignedIntToFloatDirective,
    WaitAbsDirective,
    WaitRelDirective,
)
from fprime_gds.common.loaders.ch_json_loader import ChJsonLoader
from fprime_gds.common.loaders.cmd_json_loader import CmdJsonLoader
from fprime_gds.common.loaders.prm_json_loader import PrmJsonLoader
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime_gds.common.templates.prm_template import PrmTemplate
from fprime.common.models.serialize.time_type import TimeType
from fprime.common.models.serialize.enum_type import EnumType
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.numerical_types import (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
    I16Type,
    I32Type,
    I64Type,
    I8Type,
    F32Type,
    F64Type,
    FloatType,
    IntegerType,
    NumericalType,
)
from fprime.common.models.serialize.string_type import StringType
from fprime.common.models.serialize.bool_type import BoolType
from fprime_gds.common.fpy.parser import (
    AstAnd,
    AstBoolean,
    AstComparison,
    AstElif,
    AstElifs,
    AstExpr,
    AstGetAttr,
    AstGetItem,
    AstNot,
    AstNumber,
    AstOr,
    AstReference,
    AstString,
    Ast,
    AstTest,
    AstBody,
    AstLiteral,
    AstIf,
    AstAssign,
    AstFuncCall,
    AstVar,
)
from fprime.common.models.serialize.type_base import BaseType as FppType

NUMERIC_TYPES = (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
    I16Type,
    I32Type,
    I64Type,
    I8Type,
    F32Type,
    F64Type,
)
INTEGER_TYPES = (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
    I16Type,
    I32Type,
    I64Type,
    I8Type,
)
SIGNED_INTEGER_TYPES = (
    I16Type,
    I32Type,
    I64Type,
    I8Type,
)
UNSIGNED_INTEGER_TYPES = (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
)
FLOAT_TYPES = (
    F32Type,
    F64Type,
)


# a value of type FppTypeClass is a Python `type` object representing
# the type of an Fprime value
FppTypeClass = type[FppType]


class NothingType(ABC):
    """a type which has no valid values in fprime. used to denote
    a function which doesn't return a value"""
    @classmethod
    def __subclasscheck__(cls, subclass):
        return False


# the `type` object representing the NothingType class
NothingTypeClass = type[NothingType]


class CompileException(BaseException):
    def __init__(self, msg, node: Ast):
        self.msg = msg
        self.node = node
        self.stack_trace = "\n".join(traceback.format_stack(limit=8)[:-1])

    def __str__(self):
        return f'{self.stack_trace}\nAt line {self.node.meta.line} "{self.node.node_text}": {self.msg}'


@dataclass
class FpyCallable:
    return_type: FppTypeClass | NothingTypeClass
    args: list[tuple[str, FppTypeClass]]


@dataclass
class FpyCmd(FpyCallable):
    cmd: CmdTemplate


@dataclass
class FpyMacro(FpyCallable):
    instantiate_macro: Callable[[list[FppType]], list[Directive]]
    """a function which instantiates the macro given the argument values"""


MACROS: dict[str, FpyMacro] = {
    "sleep": FpyMacro(
        NothingType, [("seconds", F64Type)], lambda args: [WaitRelDirective(int(args[0].val), int(args[0].val * 1000000) % 1000000)]
    ),
    "sleep_until": FpyMacro(
        NothingType, [("wakeup_time", TimeType)], lambda args: [WaitAbsDirective(args[0])]
    ),
    "exit": FpyMacro(NothingType, [("success", BoolType)], lambda args: [ExitDirective(args[0].val)]),
}


@dataclass
class FpyTypeCtor(FpyCallable):
    type: FppTypeClass


@dataclass
class FieldReference:
    """a reference to a field/index of an fprime type"""

    parent: "FpyReference"
    """the qualifier"""
    type: FppTypeClass
    """the fprime type of this reference"""
    offset: int
    """the constant offset in the parent type at which to find this field"""
    name: str = None
    """the name of the field, if applicable"""
    idx: int = None
    """the index of the field, if applicable"""

    def get_from(self, parent_val: FppType) -> FppType:
        """gets the field value from the parent value"""
        assert isinstance(parent_val, self.type)
        assert self.name is not None or self.idx is not None
        value = None
        if self.name is not None:
            if isinstance(parent_val, SerializableType):
                value = parent_val.val[self.name]
            elif isinstance(parent_val, TimeType):
                if self.name == "seconds":
                    value = parent_val.__secs
                elif self.name == "useconds":
                    value = parent_val.__usecs
                elif self.name == "time_base":
                    value = parent_val.__timeBase
                elif self.name == "time_context":
                    value = parent_val.__timeContext
                else:
                    assert False, self.name
            else:
                assert False, parent_val

        else:

            assert isinstance(parent_val, ArrayType), parent_val

            value = parent_val._val[self.idx]

        assert isinstance(value, self.type), (value, self.type)
        return value


# named variables can be tlm chans, prms, callables, or directly referenced consts (usually enums)
@dataclass
class FpyVariable:
    """a mutable, typed value referenced by an unqualified name"""

    type_ref: AstExpr
    """the expression denoting the var's type"""
    type: FppTypeClass | None = None
    """the resolved type of the variable. None if type unsure at the moment"""
    sreg_idx: int | None = None
    """the index of the sreg it is stored in"""


# a scope
FpyScope = dict[str, "FpyReference"]


def create_scope(
    references: dict[str, "FpyReference"],
) -> FpyScope:
    """from a flat dict of strs to references, creates a hierarchical, scoped
    dict. no two leaf nodes may have the same name"""

    base = {}

    for fqn, ref in references.items():
        names_strs = fqn.split(".")

        ns = base
        while len(names_strs) > 1:
            existing_child = ns.get(names_strs[0], None)
            if existing_child is None:
                # this scope is not defined atm
                existing_child = {}
                ns[names_strs[0]] = existing_child

            if not isinstance(existing_child, dict):
                # something else already has this name
                print(
                    f"WARNING: {fqn} is already defined as {existing_child}, tried to redefine it as {ref}"
                )
                break

            ns = existing_child
            names_strs = names_strs[1:]

        if len(names_strs) != 1:
            # broke early. skip this loop
            continue

        # okay, now ns is the complete scope of the attribute
        # i.e. everything up until the last '.'
        name = names_strs[0]

        existing_child = ns.get(name, None)

        if existing_child is not None:
            # uh oh, something already had this name with a diff value
            print(
                f"WARNING: {fqn} is already defined as {existing_child}, tried to redefine it as {ref}"
            )
            continue

        ns[name] = ref

    return base


def union_scope(lhs: FpyScope, rhs: FpyScope) -> FpyScope:
    """returns the two scopes, joined into one. if there is a conflict, chooses lhs over rhs"""
    lhs_keys = set(lhs.keys())
    rhs_keys = set(rhs.keys())
    common_keys = lhs_keys.intersection(rhs_keys)

    only_lhs_keys = lhs_keys.difference(common_keys)
    only_rhs_keys = rhs_keys.difference(common_keys)

    new = FpyScope()

    for key in common_keys:
        if not isinstance(lhs[key], dict) or not isinstance(rhs[key], dict):
            # cannot be merged cleanly. one of the two is not a scope
            print(f"WARNING: {key} is defined as {lhs[key]}, ignoring {rhs[key]}")
            new[key] = lhs[key]
            continue

        new[key] = union_scope(lhs[key], rhs[key])

    for key in only_lhs_keys:
        new[key] = lhs[key]
    for key in only_rhs_keys:
        new[key] = rhs[key]

    return new


FpyReference = (
    ChTemplate
    | PrmTemplate
    | FppType
    | FpyCallable
    | FppTypeClass
    | FpyVariable
    | FieldReference
    | dict  # FpyReference
)
"""some named concept in fpy"""


def get_ref_fpp_type_class(ref: FpyReference) -> FppTypeClass:
    """returns the fprime type of the ref, if it were to be evaluated as an expression"""
    if isinstance(ref, ChTemplate):
        result_type = ref.ch_type_obj
    elif isinstance(ref, PrmTemplate):
        result_type = ref.prm_type_obj
    elif isinstance(ref, FppType):
        # constant value
        result_type = type(ref)
    elif isinstance(ref, FpyCallable):
        # a reference to a callable isn't a type in and of itself
        # it has a return type but you have to call it (with an AstFuncCall)
        # consider making a separate "reference" type
        result_type = NothingType
    elif isinstance(ref, FpyVariable):
        result_type = ref.type
    elif isinstance(ref, type):
        # a reference to a type doesn't have a value, and so doesn't have a type,
        # in and of itself. if this were a function call to the type's ctor then
        # it would have a value and thus a type
        result_type = NothingType
    elif isinstance(ref, FieldReference):
        result_type = ref.type
    elif isinstance(ref, dict):
        # reference to a scope. scopes don't have values
        result_type = NothingType
    else:
        assert False, ref

    return result_type


@dataclass
class CompileState:
    """a collection of input, internal and output state variables and maps"""

    types: FpyScope
    """a scope whose leaf nodes are subclasses of BaseType"""
    callables: FpyScope
    """a scope whose leaf nodes are FpyCallable instances"""
    tlms: FpyScope
    """a scope whose leaf nodes are ChTemplates"""
    prms: FpyScope
    """a scope whose leaf nodes are PrmTemplates"""
    consts: FpyScope
    """a scope whose leaf nodes are instances of subclasses of BaseType"""
    variables: FpyScope = field(default_factory=dict)
    """a scope whose leaf nodes are FpyVariables"""
    runtime_values: FpyScope = None
    """a scope whose leaf nodes are tlms/prms/consts/variables, all of which
    have some value at runtime."""

    def __post_init__(self):
        self.runtime_values = union_scope(
            self.tlms,
            union_scope(self.prms, union_scope(self.consts, self.variables)),
        )

    resolved_references: dict[AstReference, FpyReference] = field(
        default_factory=dict, repr=False
    )
    """reference to its singular resolution"""

    expr_types: dict[AstExpr, FppTypeClass | NothingTypeClass] = field(
        default_factory=dict
    )
    """expr to its fprime type, or nothing type if none"""

    expr_values: dict[AstExpr, FppType | NothingType | None] = field(
        default_factory=dict
    )
    """expr to its fprime value, or nothing if no value, or None if unsure at compile time"""

    expr_registers: dict[AstExpr, int] = field(default_factory=dict)
    """expr to the register it's stored in"""

    directives: dict[Ast, list[Directive] | None] = field(default_factory=dict)
    """a list of code generated by each node, or None/empty list if no directives"""

    node_dir_counts: dict[Ast, int] = field(default_factory=dict)
    """node to the number of directives generated by it"""

    next_register: int = 0
    """the index of the next free register"""
    next_sreg: int = 0
    """the index of the next free serializable register"""

    start_line_idx: dict[Ast, int] = field(default_factory=dict)
    """the line index at which each node's directives will be included in the output"""

    errors: list[CompileException] = field(default_factory=list)
    """a list of all compile exceptions generated by passes"""

    def err(self, msg, n):
        """adds a compile exception to internal state"""
        self.errors.append(CompileException(msg, n))


class Visitor:
    """visits each class, calling a custom visit function, if one is defined, for each
    node type"""

    def _find_custom_visit_func(self, node: Ast):
        for name, func in inspect.getmembers(type(self), inspect.isfunction):
            if not name.startswith("visit") or name == "visit_default":
                # not a visitor, or the default visit func
                continue
            signature = inspect.signature(func)
            params = list(signature.parameters.values())
            assert len(params) == 3
            assert params[1].annotation is not None
            annotations = typing.get_type_hints(func)
            param_type = annotations[params[1].name]
            if isinstance(node, param_type):
                return func
        else:
            # call the default
            return type(self).visit_default

    def _visit(self, node: Ast, state: CompileState):
        visit_func = self._find_custom_visit_func(node)
        visit_func(self, node, state)

    def visit_default(self, node: Ast, state: CompileState):
        pass

    def run(self, start: Ast, state: CompileState):
        """runs the visitor, starting at the given node, descending depth-first"""

        def _descend(node: Ast):
            if not isinstance(node, Ast):
                return
            children = []
            for field in fields(node):
                field_val = getattr(node, field.name)
                if isinstance(field_val, list):
                    children.extend(field_val)
                else:
                    children.append(field_val)

            for child in children:
                if not isinstance(child, Ast):
                    continue
                _descend(child)
                if len(state.errors) != 0:
                    break
                self._visit(child, state)
                if len(state.errors) != 0:
                    break

        _descend(start)
        self._visit(start, state)


class TopDownVisitor(Visitor):

    def run(self, start: Ast, state: CompileState):
        """runs the visitor, starting at the given node, descending breadth-first"""

        def _descend(node: Ast):
            if not isinstance(node, Ast):
                return
            children = []
            for field in fields(node):
                field_val = getattr(node, field.name)
                if isinstance(field_val, list):
                    children.extend(field_val)
                else:
                    children.append(field_val)

            for child in children:
                if not isinstance(child, Ast):
                    continue
                self._visit(child, state)
                if len(state.errors) != 0:
                    break
                _descend(child)
                if len(state.errors) != 0:
                    break

        self._visit(start, state)
        _descend(start)


class AssignIds(TopDownVisitor):
    """assigns a unique id to each node to allow it to be indexed in a dict"""

    def __init__(self):
        self.next_id = 0

    def visit_default(self, node, state):
        node.id = self.next_id
        self.next_id += 1


class CreateVariables(Visitor):
    """finds all variable declarations and adds them to the variable scope"""

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        existing = state.variables.get(node.variable.var, None)
        if not existing:
            # idk what this var is. make sure it's a valid declaration
            if node.var_type is None:
                # error because this isn't an annotated assignment. right now all declarations must be annotated
                state.err(
                    "Must provide a type annotation for new variables", node.variable
                )
                return

            var = FpyVariable(node.var_type, None)
            # new var. put it in the table under this scope
            state.variables[node.variable.var] = var
            state.runtime_values[node.variable.var] = var

        if existing and node.var_type is not None:
            # redeclaring an existing variable
            state.err(f"{node.variable.var} already declared", node)
            return


class ResolveReferences(Visitor):
    """for each reference, resolve it in a specific scope based on its
    syntactic position, or fail if could not resolve"""

    def is_type_constant_size(self, type: FppTypeClass) -> bool:
        """return true if the type is statically sized"""
        if issubclass(type, StringType):
            return False

        if issubclass(type, ArrayType):
            return self.is_type_constant_size(type.MEMBER_TYPE)

        if issubclass(type, SerializableType):
            for _, arg_type, _, _ in type.MEMBER_LIST:
                if not self.is_type_constant_size(arg_type):
                    return False
            return True

        return True

    def get_attr_of_ref(
        self, parent: FpyReference, node: AstGetAttr, state: CompileState
    ) -> FpyReference | None:
        """resolve a GetAttr node relative to a given FpyReference. return the
        resolved ref, or None if none could be found. Will raise errors if not found"""

        if isinstance(parent, (FpyCallable, type)):
            # right now we don't support resolving something after a callable/type
            state.err("Invalid syntax", node)
            return None

        if isinstance(parent, dict):
            # parent is a scope
            attr = parent.get(node.attr, None)
            if attr is None:
                state.err("Unknown attribute", node)
                return None
            return attr

        # parent is a ch, prm, const, var or field

        value_type = get_ref_fpp_type_class(parent)

        assert value_type != NothingType

        if not issubclass(value_type, (SerializableType, TimeType)):
            # trying to do arr.x, but arr is not a struct
            state.err(
                "Invalid syntax (tried to access named member of a non-struct type)",
                node,
            )
            return None

        if not self.is_type_constant_size(value_type):
            state.err(
                f"{value_type} has non-constant sized members, cannot access members",
                node,
            )
            return None

        member_list: list[tuple[str, FppTypeClass]] = None
        if issubclass(value_type, SerializableType):
            member_list = [t[0:2] for t in value_type.MEMBER_LIST]
        else:
            # if it is a time type, there are some "implied" members
            member_list = []
            member_list.append(("time_base", U16Type))
            member_list.append(("time_context", U8Type))
            member_list.append(("seconds", U32Type))
            member_list.append(("useconds", U32Type))

        offset = 0
        for arg_name, arg_type in member_list:
            if arg_name == node.attr:
                return FieldReference(parent, arg_type, offset, name=arg_name)
            offset += arg_type.getMaxSize()

        state.err(f"Unknown member {node.attr}", node)
        return None

    def get_item_of_ref(
        self, parent: FpyReference, node: AstGetItem, state: CompileState
    ) -> FpyReference | None:
        """resolve a GetItem node relative to a given FpyReference. return the
        resolved ref, or None if none could be found. Will raise errors if not found"""

        if isinstance(parent, (FpyCallable, type, dict)):
            # right now we don't support resolving index after a callable/type/scope
            state.err("Invalid syntax", node)
            return None

        # parent is a ch, prm, const, var or field

        value_type = get_ref_fpp_type_class(parent)

        assert value_type != NothingType

        if not issubclass(value_type, ArrayType):
            # trying to do struct[0], but struct is not an array
            state.err(
                "Invalid syntax (tried to access indexed member of a non-array type)",
                node.item,
            )
            return None

        if not self.is_type_constant_size(value_type):
            state.err(
                f"{value_type} has non-constant sized members, cannot access members",
                node,
            )
            return None

        offset = 0
        for i in range(0, value_type.LENGTH):
            if i == node.item.value:
                return FieldReference(parent, value_type.MEMBER_TYPE, offset, idx=i)
            offset += value_type.MEMBER_TYPE.getMaxSize()

        state.err(
            f"Array access out-of-bounds (access: {node.item}, array size: {value_type.LENGTH})",
            node.item,
        )
        return None

    def resolve_if_ref(
        self, node: AstExpr, ns: FpyScope, state: CompileState
    ) -> bool:
        """if the node is a reference, try to resolve it in the given scope, and return true if success.
        otherwise, if it is not a reference, return true as it doesn't need to be resolved"""
        if not isinstance(node, AstReference):
            return True

        return self.resolve_ref_in_ns(node, ns, state) is not None

    def resolve_ref_in_ns(
        self, node: AstExpr, ns: FpyScope, state: CompileState
    ) -> FpyReference | None:
        """recursively resolves a reference in a scope, returning the resolved ref
        or none if none could be found."""
        if isinstance(node, AstVar):
            if not isinstance(ns, dict):
                return None
            ref = ns.get(node.var, None)
            if ref is None:
                return None
            state.resolved_references[node] = ref
            return ref

        parent = self.resolve_ref_in_ns(node.parent, ns, state)
        if parent is None:
            # couldn't resolve parent
            return None

        if isinstance(node, AstGetItem):
            ref = self.get_item_of_ref(parent, node, state)
            state.resolved_references[node] = ref
            return ref

        assert isinstance(node, AstGetAttr)
        ref = self.get_attr_of_ref(parent, node, state)
        state.resolved_references[node] = ref
        return ref

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        # function refs must be callables
        if not self.resolve_ref_in_ns(node.func, state.callables, state):
            state.err("Unknown callable", node.func)
            return

        for arg in node.args if node.args is not None else []:
            # arg value refs must be consts
            if not self.resolve_if_ref(arg, state.consts, state):
                state.err("Unknown const", arg)
                return

    def visit_AstIf_AstElif(self, node: AstIf | AstElif, state: CompileState):
        # if condition expr refs must be "runtime values" (tlm/prm/const/etc)
        if not self.resolve_if_ref(node.condition, state.runtime_values, state):
            state.err("Unknown runtime value", node.condition)
            return

    def visit_AstComparison(self, node: AstComparison, state: CompileState):
        # lhs/rhs side of comparison, if they are refs, must be refs to "runtime vals"
        if not self.resolve_if_ref(node.lhs, state.runtime_values, state):
            state.err("Unknown runtime value", node.lhs)
            return
        if not self.resolve_if_ref(node.rhs, state.runtime_values, state):
            state.err("Unknown runtime value", node.rhs)
            return

    def visit_AstAnd_AstOr(self, node: AstAnd | AstOr, state: CompileState):
        for val in node.values:
            if not self.resolve_if_ref(val, state.runtime_values, state):
                state.err("Unknown runtime value", val)
                return

    def visit_AstNot(self, node: AstNot, state: CompileState):
        if not self.resolve_if_ref(node.value, state.runtime_values, state):
            state.err("Unknown runtime value", node.value)
            return

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        var = self.resolve_ref_in_ns(node.variable, state.variables, state)
        if not var:
            state.err("Unknown variable", node.variable)
            return

        if node.var_type is not None:
            type = self.resolve_ref_in_ns(node.var_type, state.types, state)
            if not type:
                state.err("Unknown type", node.var_type)
                return
            var.type = type

        if not self.resolve_if_ref(node.value, state.consts, state):
            state.err("Unknown const", node.value)
            return


class CalculateExprTypes(Visitor):
    """stores in state the fprime type of each expression, or NothingType if the expr had no type"""

    def visit_AstNumber(self, node: AstNumber, state: CompileState):
        if isinstance(node.value, float):
            result_type = FloatType
        elif isinstance(node.value, int):
            result_type = IntegerType
        else:
            assert False, node.value
        state.expr_types[node] = result_type

    def visit_AstString(self, node: AstString, state: CompileState):
        state.expr_types[node] = StringType

    def visit_AstBoolean(self, node: AstBoolean, state: CompileState):
        state.expr_types[node] = BoolType

    def visit_AstReference(self, node: AstReference, state: CompileState):
        ref = state.resolved_references[node]
        state.expr_types[node] = get_ref_fpp_type_class(ref)

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        ref = state.resolved_references[node.func]
        assert isinstance(ref, FpyCallable)
        state.expr_types[node] = ref.return_type

    def visit_AstOr_AstAnd_AstNot_AstComparison(
        self, node: AstOr | AstAnd | AstNot | AstComparison, state: CompileState
    ):
        state.expr_types[node] = BoolType

    def visit_default(self, node, state):
        # coding error, missed an expr
        assert not isinstance(node, AstExpr), node


class CheckAndResolveArgumentTypes(Visitor):
    """for each syntactic node with arguments (ands/ors/nots/cmps/funcs), check that the argument
    types are right"""

    def is_literal_convertible_to(
        self, node: AstLiteral, to_type: FppTypeClass, state: CompileState
    ) -> bool:

        if isinstance(node, AstBoolean):
            return to_type == BoolType

        if isinstance(node, AstString):
            return issubclass(to_type, StringType)

        if isinstance(node, AstNumber):
            if isinstance(node.value, float):
                return issubclass(to_type, FloatType)
            if isinstance(node.value, int):
                # int literal can be converted into float or int
                return issubclass(to_type, (FloatType, IntegerType))

            assert False, node.value

        assert False, node

    def visit_AstComparison(self, node: AstComparison, state: CompileState):

        lhs_type = state.expr_types[node.lhs]
        rhs_type = state.expr_types[node.rhs]

        if not issubclass(lhs_type, NumericalType):
            state.err(f"Cannot compare non-numeric type {lhs_type}", node.lhs)
            return
        if not issubclass(rhs_type, NumericalType):
            state.err(f"Cannot compare non-numeric type {rhs_type}", node.rhs)
            return

        # args are both numeric

        # if either is generic float, pick F64. we want F64 cuz otherwise we need
        # an FPEXT to convert to F64
        if lhs_type == FloatType:
            state.expr_types[node.lhs] = F64Type
        if rhs_type == FloatType:
            state.expr_types[node.rhs] = F64Type

        if lhs_type == IntegerType and rhs_type == IntegerType:
            # use i64
            state.expr_types[node.lhs] = I64Type
            state.expr_types[node.rhs] = I64Type
            return

        if lhs_type == IntegerType:
            # try to interpret it as the rhs_type if rhs_type is integer
            if issubclass(rhs_type, IntegerType):
                state.expr_types[node.lhs] = state.expr_types[node.rhs]
            else:
                # otherwise rhs is a float. just use i64
                state.expr_types[node.lhs] = I64Type

        if rhs_type == IntegerType:
            # try to interpret it as the rhs_type if rhs_type is integer
            if issubclass(lhs_type, IntegerType):
                state.expr_types[node.rhs] = state.expr_types[node.lhs]
            else:
                # otherwise lhs is a float. just use i64
                state.expr_types[node.rhs] = I64Type

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        func = state.resolved_references[node.func]
        func_args = func.args
        node_args = node.args if node.args else []

        if len(node_args) < len(func_args):
            state.errors.append(
                CompileException(
                    f"Missing arguments (expected {len(func_args)} found {len(node_args)})",
                    node,
                )
            )
            return
        if len(node_args) > len(func_args):
            state.errors.append(
                CompileException(
                    f"Too many arguments (expected {len(func_args)} found {len(node_args)})",
                    node,
                )
            )
            return

        for value_expr, arg in zip(node_args, func_args):
            arg_name, arg_type = arg

            value_expr_type = state.expr_types[value_expr]

            if value_expr_type == arg_type or (
                isinstance(value_expr, AstLiteral)
                and self.is_literal_convertible_to(value_expr, arg_type, state)
            ):
                # arg type is good!
                state.expr_types[value_expr] = arg_type
                continue

            # it is not. these are not compatible
            state.errors.append(
                CompileException(
                    f"Cannot convert {value_expr} ({value_expr_type}) to {arg_type}",
                    value_expr,
                )
            )
            return

        # got thru all args successfully

    def visit_AstOr_AstAnd(self, node: AstOr | AstAnd, state: CompileState):
        # "or/and" can have as many args as you want. they all need to be bools tho
        for val in node.values:
            val_type = state.expr_types[val]
            if val_type != BoolType:
                state.err(f"Arguments to 'and'/'or' must be booleans", val)
                return
            state.expr_types[val] = BoolType

    def visit_AstNot(self, node: AstNot, state: CompileState):
        val_type = state.expr_types[node.value]
        if val_type != BoolType:
            state.err(f"Argument to 'not' must be boolean", node.value)
            return
        state.expr_types[node.value] = BoolType

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        var_type = state.resolved_references[node.variable].type
        value_type = state.expr_types[node.value]
        if var_type != value_type:
            if not (
                isinstance(node.value, AstLiteral)
                and self.is_literal_convertible_to(node.value, var_type, state)
            ):
                state.err(f"Cannot interpret {node.value} as {var_type}", node.value)
                return

        state.expr_types[node.value] = var_type

        if var_type.getMaxSize() > MAX_SERIALIZABLE_REGISTER_SIZE:
            state.err(f"{var_type} is too big to fit in a variable", node)
            return

    def visit_AstGetItem(self, node: AstGetItem, state: CompileState):
        # the node of the index number has no expression value, it's an arg
        # but only at syntax level
        state.expr_types[node.item] = NothingType


class CalculateExprValues(Visitor):
    """for each expr, try to calculate its constant value and store it in a map. stores None if no value could be
    calculated at compile time, and NothingType if the expr had no value"""

    def visit_AstLiteral(self, node: AstLiteral, state: CompileState):
        literal_type = state.expr_types[node]
        if literal_type != NothingType:
            assert (
                literal_type in NUMERIC_TYPES
                or issubclass(literal_type, StringType)
                or literal_type == BoolType
            ), literal_type
            state.expr_values[node] = literal_type(node.value)
        else:
            state.expr_values[node] = literal_type()

    def visit_AstReference(self, node: AstReference, state: CompileState):
        ref = state.resolved_references[node]

        if isinstance(ref, (ChTemplate, PrmTemplate, FpyVariable)):
            # we do not try to calculate or predict these values at compile time
            expr_value = None
        elif isinstance(ref, FieldReference):
            if isinstance(ref.parent, FppType):
                # ref to a field of a constant
                # get the field
                expr_value = ref.get_from(ref.parent)
            else:
                # ref to a field of smth else. no runtime val
                expr_value = None
        elif isinstance(ref, FppType):
            # constant value
            expr_value = ref
        elif isinstance(ref, FpyCallable):
            # a reference to a callable doesn't have a value, you have to actually
            # call the func
            expr_value = NothingType()
        elif isinstance(ref, type):
            # a reference to a type doesn't have a value, and so doesn't have a type,
            # in and of itself. if this were a function call to the type's ctor then
            # it would have a value
            expr_value = NothingType()
        elif isinstance(ref, dict):
            # a ref to a scope doesn't have a value
            expr_value = NothingType()
        else:
            assert False, ref

        assert expr_value is None or isinstance(expr_value, state.expr_types[node]), (
            expr_value,
            state.expr_types[node],
        )
        state.expr_values[node] = expr_value

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        func = state.resolved_references[node.func]
        assert isinstance(func, FpyCallable)
        # gather arg values
        arg_values = [
            state.expr_values[e] for e in (node.args if node.args is not None else [])
        ]
        unknown_value = any(v for v in arg_values if v is None)
        if unknown_value:
            state.expr_values[node] = None
            return

        if isinstance(func, FpyTypeCtor):
            # actually construct the type
            if issubclass(func.type, SerializableType):
                instance = func.type()
                # pass in args as a dict
                # t[0] is the arg name
                arg_dict = {t[0]: v for t, v in zip(func.type.MEMBER_LIST, arg_values)}
                instance._val = arg_dict
                state.expr_values[node] = instance

            elif issubclass(func.type, ArrayType):
                instance = func.type()
                instance._val = arg_values
                state.expr_values[node] = instance

            elif func.type == TimeType:
                state.expr_values[node] = TimeType(*arg_values)

            else:
                # no other FppTypeClasses have ctors
                assert False, func.return_type
        else:
            # don't try to calculate the value of this function call
            # it's something like a cmd or macro
            state.expr_values[node] = None

    def visit_AstTest(self, node: AstTest, state: CompileState):
        # we do not calculate compile time value of or/and/nots/cmps at the moment
        state.expr_values[node] = None

    def visit_default(self, node, state):
        # coding error, missed an expr
        assert not isinstance(node, AstExpr), node


class GenerateVariableDirectives(Visitor):
    """for each variable assignment or declaration, check the rhs was known
    at compile time, and generate a directive"""

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        existing_var = state.resolved_references[node.variable]
        # should already have been put in var table
        assert existing_var is not None

        # we should have type info about the variable
        assert existing_var.type is not None

        value_type = state.expr_types[node.value]
        value = state.expr_values[node.value]

        # already type checked
        assert value_type == type(value), (value_type, type(value))

        if value is None:
            # expr value is unknown at this point in compile
            state.err(
                f"Cannot assign {node.variable.var}: {existing_var.type} to {node.value}, as its value was not known at compile time",
                node.value,
            )
            return

        sreg_idx = existing_var.sreg_idx
        if sreg_idx is None:
            # doesn't have an sreg idx, allocate one
            sreg_idx = state.next_sreg
            state.next_sreg += 1
            existing_var.sreg_idx = sreg_idx
        val_bytes = value.serialize()
        assert len(val_bytes) == value.getMaxSize(), (
            len(val_bytes),
            value.getMaxSize(),
            value,
        )
        assert len(val_bytes) <= MAX_SERIALIZABLE_REGISTER_SIZE, len(val_bytes)
        state.directives[node] = [SetSerRegDirective(sreg_idx, val_bytes)]


class GenerateConstCmdDirectives(Visitor):
    """for each command or macro whose arguments were const at runtime (should be all at the moment),
    generate a directive"""
    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        func = state.resolved_references[node.func]
        if isinstance(func, FpyCmd):
            arg_bytes = bytes()
            for arg_node in node.args if node.args is not None else []:
                arg_value = state.expr_values[arg_node]
                if arg_value is None:
                    state.err(
                        f"Only constant arguments to commands are allowed", arg_node
                    )
                    return
                arg_bytes += arg_value.serialize()
            state.directives[node] = [CmdDirective(func.cmd.get_op_code(), arg_bytes)]
        elif isinstance(func, FpyMacro):
            arg_values = []
            for arg_node in node.args if node.args is not None else []:
                arg_value = state.expr_values[arg_node]
                if arg_value is None:
                    state.err(
                        f"Only constant arguments to macros are allowed", arg_node
                    )
                    return
                arg_values.append(arg_value)

            state.directives[node] = func.instantiate_macro(arg_values)
        else:
            state.directives[node] = None


def put_sreg_in_nreg(
    sreg_idx: int, sreg_offset: int, nreg_idx: int, size: int
) -> list[Directive]:
    if size > 4:
        return [DeserSerReg8Directive(sreg_idx, sreg_offset, nreg_idx)]
    elif size > 2:
        return [DeserSerReg4Directive(sreg_idx, sreg_offset, nreg_idx)]
    elif size > 1:
        return [DeserSerReg2Directive(sreg_idx, sreg_offset, nreg_idx)]
    elif size == 1:
        return [DeserSerReg1Directive(sreg_idx, sreg_offset, nreg_idx)]
    else:
        assert False, size


class AssignExprRegisters(Visitor):
    """assign each expr a unique register"""
    def visit_AstExpr(self, node: AstExpr, state: CompileState):
        state.expr_registers[node] = state.next_register
        state.next_register += 1


class GenerateConstExprDirectives(Visitor):
    """for each expr with a constant compile time value, generate
    directives for how to put it in its register"""
    def visit_AstExpr(self, node: AstExpr, state: CompileState):
        expr_type = state.expr_types[node]

        if node in state.directives:
            # already have directives associated with this node
            return

        if expr_type == NothingType:
            # impossible. nothing type has no value
            state.directives[node] = None
            return

        if expr_type.getMaxSize() > 8:
            # bigger than 8 bytes
            # impossible. can't fit in a register
            state.directives[node] = None
            return

        # okay, it is not nothing and it is smaller than 8 bytes.
        # should be able to put it in a reg

        register = state.expr_registers[node]

        expr_value = state.expr_values[node]

        if expr_value is None:
            # no const value
            return

        # it has a constant value at compile time
        serialized_expr_value = expr_value.serialize()
        assert len(serialized_expr_value) <= 8, len(serialized_expr_value)
        val_as_i64_bytes = bytes(8 - len(serialized_expr_value))
        val_as_i64_bytes += serialized_expr_value

        # reinterpret as an I64
        val_as_i64 = I64Type()
        val_as_i64.deserialize(val_as_i64_bytes, 0)

        state.directives[node] = [SetRegDirective(register, val_as_i64.val)]


class GenerateNonConstExprDirectives(Visitor):
    """for each expr whose value is not known at compile time, but can be calculated at run time,
    generate directives to calculate the value and put it in its register"""

    def visit_AstReference(self, node: AstReference, state: CompileState):
        if node in state.directives:
            # already know how to put it in reg, or it is impossible
            return

        expr_type = state.expr_types[node]
        ref = state.resolved_references[node]

        directives = []

        # does not have a constant compile time value

        # all references that don't have a compile time value have to go into an sreg first
        # and then into an nreg

        sreg_idx = None

        offset = 0

        base_ref = ref

        # if it's a field ref, find the parent and the offset in the parent
        while isinstance(base_ref, FieldReference):
            offset += base_ref.offset
            base_ref = base_ref.parent

        if isinstance(base_ref, FpyVariable):
            # already in an sreg
            sreg_idx = base_ref.sreg_idx
        else:
            sreg_idx = state.next_sreg
            state.next_sreg += 1

            if isinstance(base_ref, ChTemplate):
                tlm_time_sreg_idx = state.next_sreg
                state.next_sreg += 1
                directives.append(
                    GetTlmDirective(sreg_idx, tlm_time_sreg_idx, base_ref.get_id())
                )

            elif isinstance(base_ref, PrmTemplate):
                directives.append(GetPrmDirective(sreg_idx, base_ref.get_id()))

            else:
                assert (
                    False
                ), base_ref  # ref should either be impossible to put in a reg or should have a compile time val

        # pull from sreg into nreg
        directives.extend(
            put_sreg_in_nreg(
                sreg_idx, offset, state.expr_registers[node], expr_type.getMaxSize()
            )
        )

        state.directives[node] = directives

    def visit_AstAnd_AstOr(self, node: AstAnd | AstOr, state: CompileState):
        if node in state.directives:
            # already know how to put it in reg, or know that it's impossible
            return

        expr_reg = state.expr_registers[node]
        directives = []

        registers_to_compare = []
        for arg_value_expr in node.values:
            arg_value_dirs = state.directives[arg_value_expr]
            assert arg_value_dirs is not None
            directives.extend(arg_value_dirs)
            registers_to_compare.append(state.expr_registers[arg_value_expr])

        assert len(registers_to_compare) >= 2, len(registers_to_compare)

        # okay, now we have to "or" or "and" together all of the registers
        # "or/and" the first two together, put in res.
        # from then on, "or/and" the next with res

        dir_type = OrDirective if isinstance(node, AstOr) else AndDirective

        directives.append(
            dir_type(registers_to_compare[0], registers_to_compare[1], expr_reg)
        )

        for i in range(2, len(registers_to_compare)):
            directives.append(dir_type(expr_reg, registers_to_compare[i], expr_reg))

        state.directives[node] = directives

    def visit_AstNot(self, node: AstNot, state: CompileState):
        if node in state.directives:
            # already know how to put it in reg
            return

        expr_reg = state.expr_registers[node]
        directives = []
        arg_value_dirs = state.directives[node.value]
        assert arg_value_dirs is not None
        directives.extend(arg_value_dirs)
        directives.append(NotDirective(state.expr_registers[node.value], expr_reg))

        state.directives[node] = directives

    def visit_AstComparison(self, node: AstComparison, state: CompileState):
        if node in state.directives:
            # already know how to put it in reg
            return

        directives = []

        lhs_type = state.expr_types[node.lhs]
        rhs_type = state.expr_types[node.rhs]

        lhs_reg = state.expr_registers[node.lhs]
        rhs_reg = state.expr_registers[node.rhs]
        res_reg = state.expr_registers[node]

        directives.extend(state.directives[node.lhs])
        directives.extend(state.directives[node.rhs])

        fp = False
        if issubclass(lhs_type, FloatType) or issubclass(rhs_type, FloatType):
            fp = True

        if fp:
            # need to convert both lhs and rhs into F64
            # modify them in place

            # convert int to float
            if issubclass(lhs_type, IntegerType):
                if lhs_type in UNSIGNED_INTEGER_TYPES:
                    directives.append(UnsignedIntToFloatDirective(lhs_reg, lhs_reg))
                else:
                    directives.append(SignedIntToFloatDirective(lhs_reg, lhs_reg))
            if issubclass(rhs_type, IntegerType):
                if rhs_type in UNSIGNED_INTEGER_TYPES:
                    directives.append(UnsignedIntToFloatDirective(rhs_reg, rhs_reg))
                else:
                    directives.append(SignedIntToFloatDirective(rhs_reg, rhs_reg))

            # convert F32 to F64
            if lhs_type == F32Type:
                directives.append(FloatExtendDirective(lhs_reg, lhs_reg))
            if rhs_type == F32Type:
                directives.append(FloatExtendDirective(rhs_reg, rhs_reg))

        if node.op.value == "==":
            if fp:
                directives.append(FloatEqualDirective(lhs_reg, rhs_reg, res_reg))
            else:
                directives.append(IntEqualDirective(lhs_reg, rhs_reg, res_reg))
        elif node.op.value == "!=":
            if fp:
                directives.append(FloatNotEqualDirective(lhs_reg, rhs_reg, res_reg))
            else:
                directives.append(IntNotEqualDirective(lhs_reg, rhs_reg, res_reg))
        else:

            if fp:
                dir_type = FLOAT_INEQUALITY_DIRECTIVES[node.op.value]
            else:
                # if either is signed, consider both as signed
                signed = (
                    lhs_type in SIGNED_INTEGER_TYPES or rhs_type in SIGNED_INTEGER_TYPES
                )

                if signed:
                    dir_type = INT_SIGNED_INEQUALITY_DIRECTIVES[node.op.value]
                else:
                    dir_type = INT_UNSIGNED_INEQUALITY_DIRECTIVES[node.op.value]

            directives.append(dir_type(lhs_reg, rhs_reg, res_reg))

        state.directives[node] = directives


class CountNodeDirectives(Visitor):
    """count the number of directives that will be generated by each node"""

    def visit_AstIf(self, node: AstIf, state: CompileState):
        count = 0
        # include the condition
        count += state.node_dir_counts[node.condition]
        # include if stmt
        count += 1
        # include body
        count += state.node_dir_counts[node.body]
        # include a goto end of if
        count += 1

        if node.elifs is not None:
            count += state.node_dir_counts[node.elifs]
        if node.els is not None:
            count += state.node_dir_counts[node.els]

        state.node_dir_counts[node] = count

    def visit_AstElifs(self, node: AstElifs, state: CompileState):
        count = 0
        for case in node.cases:
            count += state.node_dir_counts[case]

        state.node_dir_counts[node] = count

    def visit_AstElif(self, node: AstElif, state: CompileState):
        count = 0
        # include the condition
        count += state.node_dir_counts[node.condition]
        # include if stmt
        count += 1
        # include body
        count += state.node_dir_counts[node.body]
        # include a goto end of if
        count += 1

        state.node_dir_counts[node] = count

    def visit_AstBody(self, node: AstBody, state: CompileState):
        count = 0
        for stmt in node.stmts:
            count += state.node_dir_counts[stmt]

        state.node_dir_counts[node] = count

    def visit_default(self, node, state):
        state.node_dir_counts[node] = (
            len(state.directives[node]) if state.directives.get(node) is not None else 0
        )


class CalculateStartLineIdx(TopDownVisitor):
    """based on the number of directives generated by each node, calculate the start line idx
    of each node's directives"""
    def visit_AstBody(self, node: AstBody, state: CompileState):
        if node not in state.start_line_idx:
            state.start_line_idx[node] = 0

        start_idx = state.start_line_idx[node]

        line_idx = start_idx
        for stmt in node.stmts:
            state.start_line_idx[stmt] = line_idx
            line_idx += state.node_dir_counts[stmt]

    def visit_AstIf(self, node: AstIf, state: CompileState):
        line_idx = state.start_line_idx[node]
        state.start_line_idx[node.condition] = line_idx
        line_idx += state.node_dir_counts[node.condition]
        # include if stmt
        line_idx += 1
        state.start_line_idx[node.body] = line_idx
        line_idx += state.node_dir_counts[node.body]
        # include goto stmt
        line_idx += 1
        if node.elifs is not None:
            state.start_line_idx[node.elifs] = line_idx
            line_idx += state.node_dir_counts[node.elifs]
        if node.els is not None:
            state.start_line_idx[node.els] = line_idx
            line_idx += state.node_dir_counts[node.els]

    def visit_AstElifs(self, node: AstElifs, state: CompileState):
        line_idx = state.start_line_idx[node]
        for case in node.cases:
            state.start_line_idx[case] = line_idx
            line_idx += state.node_dir_counts[case]

    def visit_AstElif(self, node: AstElif, state: CompileState):
        line_idx = state.start_line_idx[node]
        state.start_line_idx[node.condition] = line_idx
        line_idx += state.node_dir_counts[node.condition]
        # include if dir
        line_idx += 1
        state.start_line_idx[node.body] = line_idx
        line_idx += state.node_dir_counts[node.body]
        # include a goto end of if
        line_idx += 1


class GenerateBodyDirectives(Visitor):
    """concatenate all directives together for each AstBody"""

    def visit_AstIf(self, node: AstIf, state: CompileState):
        start_line_idx = state.start_line_idx[node]

        all_dirs = []

        cases: list[tuple[AstExpr, AstBody]] = []
        goto_ends: list[GotoDirective] = []

        cases.append((node.condition, node.body))

        if node.elifs is not None:
            for case in node.elifs.cases:
                cases.append((case.condition, case.body))

        for case in cases:
            case_dirs = []
            # include the condition
            case_dirs.extend(state.directives[case[0]])
            # include if stmt (update the end idx later)
            if_dir = IfDirective(state.expr_registers[case[0]], -1)

            case_dirs.append(if_dir)
            # include body
            case_dirs.extend(state.directives[case[1]])
            # include a temporary goto end of if, will be refined later
            goto_dir = GotoDirective(-1)
            case_dirs.append(goto_dir)
            goto_ends.append(goto_dir)

            # if false, skip the body and goto
            if_dir.false_goto_stmt_index = (
                start_line_idx + len(all_dirs) + len(case_dirs)
            )

            all_dirs.extend(case_dirs)

        if node.els is not None:
            all_dirs.extend(state.directives[node.els])

        for goto in goto_ends:
            goto.statement_index = start_line_idx + len(all_dirs)

        state.directives[node] = all_dirs

    def visit_AstBody(self, node: AstBody, state: CompileState):
        dirs = []
        for stmt in node.stmts:
            stmt_dirs = state.directives.get(stmt, None)
            if stmt_dirs is not None:
                dirs.extend(stmt_dirs)

        state.directives[node] = dirs


def get_base_compile_state(dictionary: str) -> CompileState:
    """return the initial state of the compiler, based on the given dict path"""
    cmd_json_dict_loader = CmdJsonLoader(dictionary)
    (cmd_id_dict, cmd_name_dict, versions) = cmd_json_dict_loader.construct_dicts(
        dictionary
    )

    ch_json_dict_loader = ChJsonLoader(dictionary)
    (ch_id_dict, ch_name_dict, versions) = ch_json_dict_loader.construct_dicts(
        dictionary
    )
    prm_json_dict_loader = PrmJsonLoader(dictionary)
    (prm_id_dict, prm_name_dict, versions) = prm_json_dict_loader.construct_dicts(
        dictionary
    )
    # the type name dict is a mapping of a fully qualified name to an fprime type
    # here we put into it all types found while parsing all cmds, params and tlm channels
    type_name_dict: dict[str, FppTypeClass] = cmd_json_dict_loader.parsed_types
    type_name_dict.update(ch_json_dict_loader.parsed_types)
    type_name_dict.update(prm_json_dict_loader.parsed_types)

    # enum const dict is a dict of fully qualified enum const name (like Ref.Choice.ONE) to its fprime value
    enum_const_name_dict: dict[str, FppType] = {}

    # find each enum type, and put each of its values in the enum const dict
    for name, typ in type_name_dict.items():
        if issubclass(typ, EnumType):
            for enum_const_name, val in typ.ENUM_DICT.items():
                enum_const_name_dict[name + "." + enum_const_name] = typ(
                    enum_const_name
                )

    # insert the implicit types into the dict
    type_name_dict["Fw.Time"] = TimeType
    for typ in NUMERIC_TYPES:
        type_name_dict[typ.get_canonical_name()] = typ
    type_name_dict["bool"] = BoolType
    # note no string type at the moment

    callable_name_dict: dict[str, FpyCallable] = {}
    # add all cmds to the callable dict
    for name, cmd in cmd_name_dict.items():
        cmd: CmdTemplate
        args = []
        for arg_name, _, arg_type in cmd.arguments:
            args.append((arg_name, arg_type))
        # cmds are thought of as callables with a "NothingType" return value
        callable_name_dict[name] = FpyCmd(NothingType, args, cmd)

    # for each type in the dict, if it has a constructor, create an FpyTypeCtor
    # object to track the constructor and put it in the callable name dict
    for name, typ in type_name_dict.items():
        args = []
        if issubclass(typ, SerializableType):
            for arg_name, arg_type, _, _ in typ.MEMBER_LIST:
                args.append((arg_name, arg_type))
        elif issubclass(typ, ArrayType):
            for i in range(0, typ.LENGTH):
                args.append(("e" + str(i), typ.MEMBER_TYPE))
        elif issubclass(typ, TimeType):
            args.append(("time_base", U16Type))
            args.append(("time_context", U8Type))
            args.append(("seconds", U32Type))
            args.append(("useconds", U32Type))
        else:
            # bool, enum, string or numeric type
            # none of these have callable ctors
            continue

        callable_name_dict[name] = FpyTypeCtor(typ, args, typ)

    # for each macro function, add it to the callable dict
    for macro_name, macro in MACROS.items():
        callable_name_dict[macro_name] = macro

    state = CompileState(
        tlms=create_scope(ch_name_dict),
        prms=create_scope(prm_name_dict),
        types=create_scope(type_name_dict),
        callables=create_scope(callable_name_dict),
        consts=create_scope(enum_const_name_dict),
    )
    return state


def compile(body: AstBody, dictionary: str) -> list[Directive]:
    state = get_base_compile_state(dictionary)
    passes: list[Visitor] = [
        AssignIds(),
        # based on assignment syntax nodes, we know which variables exist where
        CreateVariables(),
        # now that variables have been defined, all names/attributes/indices (references)
        # should be defined
        ResolveReferences(),
        # now that we know what all refs point to, we should be able to figure out the type
        # of every expression
        CalculateExprTypes(),
        # now that we know the type of each expr, we can type check all function calls
        # and also narrow down ambiguous argument types
        CheckAndResolveArgumentTypes(),
        # okay, now that we're sure we're passing in all the right args to each func,
        # we can calculate values of type ctors etc etc
        CalculateExprValues(),
        # now that we know variable values, we can generate directives for vars
        GenerateVariableDirectives(),
        # give each expr its own register
        AssignExprRegisters(),
        # for cmds, which have constant arguments, generate the corresponding directives
        GenerateConstCmdDirectives(),
        # for expressions which have constant values, generate corresponding directives
        # to put the expr in its register
        GenerateConstExprDirectives(),
        # for expressions which don't have constant values, generate directives to
        # calculate the expr at runtime and put it in its register
        GenerateNonConstExprDirectives(),
        # count the number of directives generated by each node
        CountNodeDirectives(),
        # calculate the index that the node will correspond to in the output file
        CalculateStartLineIdx(),
        # generate directives for each body node, including the root
        GenerateBodyDirectives(),
    ]

    for compile_pass in passes:
        compile_pass.run(body, state)
        for error in state.errors:
            raise error

    return state.directives[body]



import platform
import tempfile
import unittest
from pathlib import Path
from unittest import mock

from fprime_gds.executables import run_deployment


class TestRunDeployment(unittest.TestCase):

    def test_as_in_installation_instructions(self):
        # Same as the "Testing F´ GDS Installation Via Running HTML GUI" from
        # https://nasa.github.io/fprime/INSTALL.html
        # fprime-gds -g html -d <path to fprime checkout>/Ref/build-artifacts/<platform>/Test
        with tempfile.TemporaryDirectory() as temporary_directory:
            self.create_fake_deployment_structure(temporary_directory)
            with mock.patch("sys.argv", ["main", "-g", "html", "-d", str(Path(temporary_directory) / platform.system() / "Test")]):
                run_deployment.parse_args()

    def create_fake_deployment_structure(self, temporary_directory):
        system_dir = Path(temporary_directory) / platform.system() / "Test"

        bin_dir = system_dir / "bin"
        bin_dir.mkdir(parents=True)
        bin_file = bin_dir / "Test"
        with bin_file.open(mode="wb") as fake_app:
            fake_app.write("fake app".encode("utf-8"))

        dictionary_dir = system_dir / "dict"
        dictionary_dir.mkdir(parents=True)
        dictionary_file = dictionary_dir / "TestTopologyAppDictionary.xml"
        with dictionary_file.open(mode="w") as fake_dictionary:
            fake_dictionary.write("<test></test>")



"""
predicates.py:

This file contains basic predicates as well as event and telemetry predicates used by the
gds_test_api.py. The predicates are organized by type and can be used to search histories.

:author: koran
"""
from inspect import signature

from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.data_types.event_data import EventData


##########################################################################################
# Base class/helpers
##########################################################################################
class predicate:
    def __call__(self, item):
        """
        A parent class to ensure that all predicates are callable, and return readable strings
        """
        raise NotImplementedError(
            "This predicate did not override __call__(self, object)"
        )

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        raise NotImplementedError("This predicate did not override __str__(self)")

    def __repr__(self):
        """
        Unless manually overwritten by a subclass both repr and str will return the same string.
        """
        return str(self)


def is_predicate(pred):
    """
    a helper function to determine if an object can be used as a predicate.

    :return: a boolean value of whether the function is a predicate instance or has
        both __str__ and __call__ methods.
    """
    if isinstance(pred, predicate):
        return True
    if callable(pred):
        sig = signature(pred.__call__)
        arg_count = len(sig.parameters)
        if arg_count == 1 and hasattr(pred, "__str__"):
            return True
    return False


def get_descriptive_string(value, pred_function):
    """
    a helper function that formats a predicate and argument in a nice human-readable format

    Args:
        value: the argument of the predicate
        pred_function: a predicate function
    """
    return f"F({value}), where F(x) evaluates\n\t {pred_function}"


##########################################################################################
# Comparison predicates
##########################################################################################
class less_than(predicate):
    def __init__(self, value):
        """
        A predicate that evaluates a less-than comparison
        :param value: To return true, the predicate must be called on an object that is less
            than this value
        """

        self.upper_limit = value

    def __call__(self, actual):
        """

        :param actual: the value to compare
        """
        try:
            return actual < self.upper_limit
        except TypeError:
            return False

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x < {self.upper_limit}"


class greater_than(predicate):
    def __init__(self, value):
        """
        A predicate that evaluates a greater-than comparison
        :param value: To return true, the predicate must be called on an object that is less
            than this value
        """
        self.lower_limit = value

    def __call__(self, actual):
        """

        :param actual: the value to compare
        """
        try:
            return actual > self.lower_limit
        except TypeError:
            return False

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x > {self.lower_limit}"


class equal_to(predicate):
    def __init__(self, value):
        """
        A predicate that evaluates an equivalent comparison
        :param value: To return true, the predicate must be called on an object that is equal
            to this value
        """
        self.expected = value

    def __call__(self, actual):
        """

        :param actual: the value to compare
        """
        try:
            return actual == self.expected
        except TypeError:
            return False

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x == {self.expected}"


class not_equal_to(predicate):
    def __init__(self, value):
        """
        A predicate that evaluates a not-equivalent comparison
        :param value: To return true, the predicate must be called on an object that is not
            equal to this value
        """
        self.expected = value

    def __call__(self, actual):
        """

        :param actual: the value to compare
        """
        try:
            return actual != self.expected
        except TypeError:
            return False

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x != {self.expected}"


class less_than_or_equal_to(predicate):
    def __init__(self, value):
        """
        A predicate that evaluates a less-than-or-equals comparison
        :param value: To return true, the predicate must be called on an object that is less
            than or equal to this value
        """
        self.upper_limit = value

    def __call__(self, actual):
        """

        :param actual: the value to compare
        """
        try:
            return actual <= self.upper_limit
        except TypeError:
            return False

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x <= {self.upper_limit}"


class greater_than_or_equal_to(predicate):
    def __init__(self, value):
        """
        A predicate that evaluates greater-than-or-equals comparison
        :param value: To return true, the predicate must be called on an object that is
            greater than or equal to this value
        """
        self.lower_limit = value

    def __call__(self, actual):
        """

        :param actual: the value to compare
        """
        try:
            return actual >= self.lower_limit
        except TypeError:
            return False

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x >= {self.lower_limit}"


class within_range(predicate):
    def __init__(self, lower, upper):
        """
        A predicate that evaluates if the argument is between the two values
        :param lower: To return true, the predicate must be called on an object that is
            greater than this value
        :param upper: To return true, the predicate must be called on an object that is less
            than this value
        """
        self.upper_limit = upper
        self.lower_limit = lower

    def __call__(self, actual):
        """

        :param actual: the value to evaluate
        """
        try:
            return self.lower_limit <= actual <= self.upper_limit
        except TypeError:
            return False

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"{self.lower_limit} <= x <= {self.upper_limit}"


##########################################################################################
# Set predicates
##########################################################################################
class is_a_member_of(predicate):
    def __init__(self, collection):
        """
        Constructs a predicate that evaluates if the argument is equivalent to any member in the set
        :param collection: To return true, the predicate must be called on an object that is
            equivalent to any object in this list
        """
        self.set = collection

    def __call__(self, item):
        """
        Evaluates the predicate

        :param item: the object to search for then evaluate
        """
        return any(item == x for x in self.set)

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x ∈ {self.set}"


class is_not_a_member_of(predicate):
    def __init__(self, collection):
        """
        Constructs a predicate that evaluates if the argument is not equivalent to all members in
        the set
        :param collection: To return true, the predicate must be called on an object that is
            not equivalent to any object in this collection
        """
        self.set = collection

    def __call__(self, item):
        """

        :param item: the object to search for then evaluate
        """
        return all(item != x for x in self.set)

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x ∉ {self.set}"


##########################################################################################
# Logic predicates
##########################################################################################
class always_true(predicate):
    def __call__(self, obj):
        """
        used as a placeholder by other predicates. This is like a logical TRUE signal.
        :param obj: the object or value to evaluate
        """
        return True

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return "True"


class invert(predicate):
    def __init__(self, pred):
        """
        A predicate that negates a given predicate. This predicate can be used like a NOT
        gate when combining predicates.
        :param pred: The predicate to be negated.
        """
        if is_predicate(pred):
            self.pred = pred

    def __call__(self, item):
        """

        :param item: the object or value to evaluate
        """
        return not self.pred(item)

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"˜({self.pred})."


class satisfies_all(predicate):
    def __init__(self, pred_list):
        """
        A predicate that evaluates if the argument satisfies all predicates in the given list.
        This predicate can be used like an AND gate of N elements when combining predicates.
        :param pred_list: a list of predicates
        """
        self.p_list = [pred for pred in pred_list if is_predicate(pred)]

    def __call__(self, item):
        """

        :param item: the object or value to evaluate
        """
        return all(pred(item) for pred in self.p_list)

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"∀ P ∈ A : P(x) is True, when A is {self.p_list}"


class satisfies_any(predicate):
    def __init__(self, pred_list):
        """
        A predicate that evaluates if the argument satisfies any predicate in the given list.
        This predicate can be used like an OR gate of N elements when combining predicates.
        :param pred_list: a list of predicates
        """
        self.p_list = [pred for pred in pred_list if is_predicate(pred)]

    def __call__(self, item):
        """

        :param item: the object or value to evaluate
        """
        return any(pred(item) for pred in self.p_list)

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"∃ P ∈ A : P(x) is True, when A is {self.p_list}"


##########################################################################################
# Test API predicates
##########################################################################################
class args_predicate(predicate):
    def __init__(self, args):
        """
        A predicate for evaluating argument fields. Arguments can be specified by value, by
        predicate or as don't care (None). By inserting None into the argument list, args_predicate
        will accept any response given for that argument index.

        Args:
            args: a list of expected arguments (list of values, predicates, and None)
        """
        self.arg_spec = []
        ignored = always_true()
        if not isinstance(args, (list, tuple)):
            args = [args]
        for arg in args:
            if arg is None:
                self.arg_spec.append(ignored)
            elif is_predicate(arg):
                self.arg_spec.append(arg)
            else:
                self.arg_spec.append(equal_to(arg))

    def __call__(self, actual):
        """
        Determines if the given array satisfies the given argument specification.
        """
        if not isinstance(actual, (list, tuple)):
            actual = [actual]
        if len(actual) != len(self.arg_spec):
            return False
        return all(self.arg_spec[i](actual[i]) for i in range(len(self.arg_spec)))

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"True IFF ∀ pi ∈ P and xi ∈ x; pi(xi) is True. Where P is {self.arg_spec}"


class event_predicate(predicate):
    def __init__(
        self, id_pred=None, args_pred=None, severity_pred=None, time_pred=None
    ):
        """
        A predicate for specifying an EventData object from data_types.event_data. This predicate
        can be used to search a history. If arguments passed into this constructor are not
        subclasses of predicate, they will be ignored. If an argument is unspecified, the predicate
        will ignore that field when evaluating an EventData object.

        Args:
            id_pred: optional predicate to call on the EventData instance's id field
            args_pred: optional predicate to call on a list of the EventData 's argument values
            severity_pred: optional predicate to call on the EventData instances severity field
            time_pred: optional predicate to call on the EventData instance's timestamp
        """
        true_pred = always_true()
        self.id_pred = true_pred
        self.args_pred = true_pred
        self.severity_pred = true_pred
        self.time_pred = true_pred
        if is_predicate(id_pred):
            self.id_pred = id_pred
        if is_predicate(args_pred):
            self.args_pred = args_pred
        if is_predicate(severity_pred):
            self.severity_pred = severity_pred
        if is_predicate(time_pred):
            self.time_pred = time_pred

    def __call__(self, event):
        """
        The event_predicate checks that the telemetry object is an instance of EventData and will
        raise an error if the check fails. Then event_predicate will evaluate whether event's
        EventData fields satisfy the id_pred, value_pred, and time_pred specified.

        Args:
            event: an instance of EventData (object)
        """
        if not isinstance(event, EventData):
            return False
        if (
            self.id_pred(event.get_id())
            and self.time_pred(event.get_time())
            and self.severity_pred(event.get_severity())
        ):
            args = [arg.val for arg in event.get_args()]
            if self.args_pred(args):
                return True
        return False

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        msg = "True IFF: x is an EventData object"
        if not isinstance(self.id_pred, always_true):
            msg += f" and x's id satisfies ({self.id_pred})"
        if not isinstance(self.args_pred, always_true):
            msg += f" and x's args satisfy ({self.args_pred})"
        if not isinstance(self.severity_pred, always_true):
            msg += f" and x's severity satisfies ({self.severity_pred})"
        if not isinstance(self.time_pred, always_true):
            msg += f" and x's time satisfies ({self.time_pred})"
        return msg


class telemetry_predicate(predicate):
    def __init__(self, id_pred=None, value_pred=None, time_pred=None):
        """
        A predicate for specifying a ChData object from data_types.ch_data. This predicate can be
        used to search a history. If arguments passed into this constructor are not subclasses of
        predicate, they will be ignored.If an argument is unspecified, the predicate will ignore
        that field when evaluating an EventData object.

        Args:
            id_pred: optional predicate to call on the ChData instance's id field
            value_pred: optional predicate to call on the ChData instance's id field
            time_pred: optional predicate to call on the ChData instance's timestamp
        """
        true_pred = always_true()
        self.id_pred = true_pred
        self.value_pred = true_pred
        self.time_pred = true_pred
        if is_predicate(id_pred):
            self.id_pred = id_pred
        if is_predicate(value_pred):
            self.value_pred = value_pred
        if is_predicate(time_pred):
            self.time_pred = time_pred

    def __call__(self, telemetry):
        """
        The telemetry_predicate checks that the telemetry object is an instance of ChData and will
        raise an error if the check fails. Then telemetry_predicate will evaluate whether
        telemetry's ChData fields satisfy the id_pred, value_pred and time_pred specified.

        Args:
            telemetry: an instance of ChData (object)
        """
        return (
            isinstance(telemetry, ChData)
            and self.id_pred(telemetry.get_id())
            and self.value_pred(telemetry.get_val())
            and self.time_pred(telemetry.get_time()))

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        msg = "True IFF: x is a ChData object"
        if not isinstance(self.id_pred, always_true):
            msg += f" and x's id satisfies ({self.id_pred})"
        if not isinstance(self.value_pred, always_true):
            msg += f" and x's value satisfies ({self.value_pred})"
        if not isinstance(self.time_pred, always_true):
            msg += f" and x's time satisfies ({self.time_pred})"
        return msg



"""
@brief Encoder for file data

This encoder takes in file objects, serializes them, and sends the results
to all registered senders.

Serialized command format:
    +--------------------------------+          -
    | Header = "A5A5 "               |          |
    | (5 byte string)                |          |
    +--------------------------------+      Added by
    | Destination = "GUI " or "FSW " |       Sender
    | (4 byte string)                |          |
    +--------------------------------+          -
    | Command descriptor             |
    | (0x????????)                   |
    | (4 byte number)                |
    +--------------------------------+
    | Length of descriptor, opcode,  |
    | and argument data              |
    | (4 bytes)                      |
    +--------------------------------+
    | Packet type                    |
    | (4 bytes)                      |
    +--------------------------------+
    | Sequence ID                    |
    | (4 bytes)                      |
    +--------------------------------+
    | Type specific                  |
    +--------------------------------+
    | Type specific                  |
    +--------------------------------+
    | ...                            |
    +--------------------------------+
    | Type specific                  |
    +--------------------------------+

@date Created February 2020-02-05
@author mstarch

@bug No known bugs
"""

import struct

from fprime.common.models.serialize.numerical_types import U32Type
from fprime.constants import DATA_ENCODING

from fprime_gds.common.data_types.file_data import FilePacketType
from fprime_gds.common.utils.data_desc_type import DataDescType

from . import encoder


class FileEncoder(encoder.Encoder):
    """
    Encodes the file data. This plugs into the uplink system and allows for data to be sent to the spacecraft.
    """

    def __init__(self, config=None):
        """
        Constructs a file encoder. Defaults to FSW as destination..

        Args:
            config (ConfigManager, default=None): Object with configuration data
                    for the sizes of fields in the binary data. If None passed,
                    defaults are used.
        """
        super().__init__(config)
        self.len_obj = self.config.get_type("msg_len")

    def encode_api(self, data):
        """
        Encodes specific file packets. This will allow the data to be sent out.

        :param data: FilePacket type to send.
        :return: encoded bytes data
        """
        out_data = struct.pack(">BI", int(data.packetType.value), data.seqID)
        # Packet Type determines the variables following the seqID
        if data.packetType == FilePacketType.START:
            file_src_enc = data.sourcePath.encode(DATA_ENCODING)
            file_dst_enc = data.destPath.encode(DATA_ENCODING)
            out_data += struct.pack(">IB", data.size, len(file_src_enc))
            out_data += file_src_enc
            out_data += struct.pack(">B", len(file_dst_enc))
            out_data += file_dst_enc
        elif data.packetType == FilePacketType.DATA:
            out_data += struct.pack(">IH", data.offset, data.length)
            out_data += data.dataVar
        elif data.packetType == FilePacketType.END:
            out_data += struct.pack(">I", data.hashValue)
        elif data.packetType != FilePacketType.CANCEL:
            msg = f"Invalid packet type found while encoding: {data.packetType}"
            raise Exception(msg)
        descriptor = U32Type(DataDescType["FW_PACKET_FILE"].value).serialize()
        length_obj = self.config.get_type("msg_len")
        length_obj.val = len(descriptor) + len(out_data)
        header = U32Type(0x5A5A5A5A).serialize() + length_obj.serialize() + descriptor
        return header + out_data



"""
@brief Event Template class

Instances of this class describe a specific event type. For example: AF_ASSERT_0
or cmdSeq_CS_CmdStarted

@date Created July 2, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime.common.models.serialize import type_base
from fprime.common.models.serialize.type_exceptions import TypeMismatchException

from fprime_gds.common.utils.event_severity import EventSeverity

from . import data_template


class EventTemplate(data_template.DataTemplate):
    """Class to create event templates to describe specific event types"""

    def __init__(
        self, event_id, name, component, args, severity, format_str, description=None
    ):
        """
        Constructor

        Args:
            event_id: The ID of the event being described
            name: event name as a string
            component: Component that produces the Event
            args: List of arguments in tuple form. Each tuple should be:
                  (arg name, arg description, arg obj). Where arg obj is an
                  object of a type derived from the class Base Type. Arg
                  description may be None.
            severity: event severity as an EventSeverity Enum
            format_str: Format string for the event's arguments
            description: (Optional) Event Description
        """
        super().__init__()
        # Make sure correct types are passed
        if not isinstance(event_id, int):
            raise TypeMismatchException(int, type(event_id))

        if not isinstance(name, str):
            raise TypeMismatchException(str, type(name))

        if not isinstance(component, str):
            raise TypeMismatchException(str, type(component))

        if not isinstance(format_str, str):
            raise TypeMismatchException(str, type(format_str))

        if not isinstance(args, list):
            raise TypeMismatchException(list, type(args))

        for (arg_name, arg_desc, arg_type) in args:
            if not isinstance(arg_name, str):
                raise TypeMismatchException(str, type(arg_name))

            if arg_desc is not None and not isinstance(arg_desc, str):
                raise TypeMismatchException(str, type(arg_desc))

            if not issubclass(arg_type, type_base.BaseType):
                raise TypeMismatchException(type_base.BaseType, arg_type)

        if description is not None and not isinstance(description, str):
            raise TypeMismatchException(str, type(description))

        if not isinstance(severity, EventSeverity):
            raise TypeMismatchException("EventSeverity", type(severity))

        # Initialize event internal variables
        self.id = event_id
        self.name = name
        self.comp_name = component
        self.args = args
        self.severity = severity
        self.format_str = format_str
        self.description = description

    def get_full_name(self):
        """
        Get the full name of this event

        Returns:
            The full name (component.channel) for this event
        """
        return f"{self.comp_name}.{self.name}"

    def get_id(self):
        return self.id

    def get_name(self):
        return self.name

    def get_comp_name(self):
        return self.comp_name

    def get_severity(self):
        """
        Returns the event's severity as an EventSeverity Enum.

        Returns:
            The event's severity as an EventSeverity Enum
        """
        return self.severity

    def get_format_str(self):
        return self.format_str

    def get_description(self):
        return self.description

    def get_args(self):
        """
        Returns a list of argument information

        Returns:
            A list of tuples where each tuple represents an argument. Each tuple
            in the form: (arg name, arg description, arg obj). Where arg obj is
            an object of a type derived from the class Base Type. Arg
            description may be None.
        """
        return self.args


if __name__ == "__main__":
    pass



from pathlib import Path
import tempfile
from fprime_gds.common.fpy.bytecode.directives import Directive, serialize_directives
from fprime_gds.common.fpy.codegen import compile
from fprime_gds.common.fpy.parser import parse
from fprime_gds.common.testing_fw.api import IntegrationTestAPI


def compile_seq(fprime_test_api, seq: str) -> list[Directive]:
    return compile(parse(seq), fprime_test_api.pipeline.dictionary_path)


def run_seq(fprime_test_api: IntegrationTestAPI, directives: list[Directive]):
    file = tempfile.NamedTemporaryFile(suffix=".bin", delete=False)

    serialize_directives(directives, Path(file.name))

    fprime_test_api.send_and_assert_command("ComFpy.cmdSeq.RUN", [file.name, "BLOCK"], timeout=4)


def assert_compile_success(fprime_test_api, seq: str):
    compile_seq(fprime_test_api, seq)


def assert_run_success(fprime_test_api, seq: str):
    directives = compile_seq(fprime_test_api, seq)

    run_seq(fprime_test_api, directives)


def assert_compile_failure(fprime_test_api, seq: str):
    try:
        compile_seq(fprime_test_api, seq)
    except BaseException as e:
        return
    raise RuntimeError("compile_seq succeeded")


def assert_run_failure(fprime_test_api, seq: str):
    directives = compile_seq(fprime_test_api, seq)
    try:
        run_seq(fprime_test_api, directives)
    except BaseException as e:
        return
    raise RuntimeError("run_seq succeeded")


def test_simple_var(fprime_test_api):
    seq = """
var: U32 = 1
"""

    assert_run_success(fprime_test_api, seq)

def test_float_log_literal(fprime_test_api):
    seq = """
var: F32 = 1.000e-5
"""

    assert_compile_success(fprime_test_api, seq)


def test_exit_success(fprime_test_api):
    seq = """
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_exit_failure(fprime_test_api):
    seq = """
exit(False)
"""
    assert_run_failure(fprime_test_api, seq)


def test_large_var(fprime_test_api):
    seq = """
var: Svc.DpRecord = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
"""

    assert_run_success(fprime_test_api, seq)


def test_var_wrong_rhs(fprime_test_api):
    seq = """
x: U32 = 1
var: U32 = x
"""

    assert_compile_failure(fprime_test_api, seq)


def test_nonexistent_var(fprime_test_api):
    seq = """
var = 1
"""

    assert_compile_failure(fprime_test_api, seq)


def test_create_after_assign_var(fprime_test_api):
    seq = """
var = 1
var: U32 = 2
"""

    assert_compile_failure(fprime_test_api, seq)


def test_bad_assign_type(fprime_test_api):
    seq = """
var: failure = 1
"""

    assert_compile_failure(fprime_test_api, seq)


def test_weird_assign_type(fprime_test_api):
    seq = """
var: CdhCore.cmdDisp.CMD_NO_OP = 1
"""

    assert_compile_failure(fprime_test_api, seq)


def test_reassign(fprime_test_api):
    seq = """
var: U32 = 1
var = 2
"""

    assert_run_success(fprime_test_api, seq)


def test_reassign_ann(fprime_test_api):
    seq = """
var: U32 = 1
var: U32 = 2
"""
    assert_compile_failure(fprime_test_api, seq)


def test_assign_inconsistent_type(fprime_test_api):
    seq = """
var: U32 = 1
var: U16 = 2
"""

    assert_compile_failure(fprime_test_api, seq)


def test_call_cmd(fprime_test_api):
    seq = """
CdhCore.cmdDisp.CMD_NO_OP()
"""
    assert_run_success(fprime_test_api, seq)


def test_call_cmd_with_str_arg(fprime_test_api):
    seq = """
CdhCore.cmdDisp.CMD_NO_OP_STRING("hello world")
"""
    assert_run_success(fprime_test_api, seq)


def test_call_cmd_with_int_arg(fprime_test_api):
    seq = """
FpyDemo.sendBuffComp.PARAMETER3_PRM_SET(4)
"""
    assert_run_success(fprime_test_api, seq)


def test_bad_enum_ctor(fprime_test_api):
    seq = """
FpyDemo.SG5.Settings(123, 0.5, 0.5, FpyDemo.SignalType(1))
"""
    assert_compile_failure(fprime_test_api, seq)


def test_cmd_with_enum(fprime_test_api):
    seq = """
FpyDemo.SG5.Settings(123, 0.5, 0.5, FpyDemo.SignalType.TRIANGLE)
"""
    assert_run_success(fprime_test_api, seq)


def test_instantiate_type_for_cmd(fprime_test_api):
    seq = """
FpyDemo.typeDemo.CHOICE_PAIR(FpyDemo.ChoicePair(FpyDemo.Choice.ONE, FpyDemo.Choice.TWO))
"""
    assert_run_success(fprime_test_api, seq)


def test_var_with_enum_type(fprime_test_api):
    seq = """
var: FpyDemo.Choice = FpyDemo.Choice.ONE
"""

    assert_run_success(fprime_test_api, seq)


def test_simple_if(fprime_test_api):
    seq = """
var: bool = True

# use exit(True) if we want the sequence to succeed
# exit(False) if we want it to fail. helpful for testing.

if var:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_or_expr(fprime_test_api):
    seq = """
if True or False:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_not_expr(fprime_test_api):
    seq = """
if not False:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_or_expr_with_vars(fprime_test_api):
    seq = """
var1: bool = True
var2: bool = False

if var1 or var2:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_geq(fprime_test_api):
    seq = """
if 2 >= 1:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_geq_tlm(fprime_test_api):
    seq = """
CdhCore.cmdDisp.CMD_NO_OP()
if CdhCore.cmdDisp.CommandsDispatched >= 1:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_large_elifs(fprime_test_api):
    seq = """
if CdhCore.cmdDisp.CommandsDispatched == 0:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("0")
elif CdhCore.cmdDisp.CommandsDispatched == 1:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("1")
elif CdhCore.cmdDisp.CommandsDispatched == 2:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("2")
elif CdhCore.cmdDisp.CommandsDispatched == 3:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("3")
elif CdhCore.cmdDisp.CommandsDispatched == 4:
    CdhCore.cmdDisp.CMD_NO_OP_STRING("4")
else:
    CdhCore.cmdDisp.CMD_NO_OP_STRING(">4")
"""

    assert_run_success(fprime_test_api, seq)


def test_int_as_stmt(fprime_test_api):
    seq = """
2
"""

    assert_compile_failure(fprime_test_api, seq)


def test_complex_as_stmt(fprime_test_api):
    seq = """
CdhCore.cmdDisp.CMD_NO_OP
"""

    assert_compile_failure(fprime_test_api, seq)


def test_get_struct_member(fprime_test_api):
    seq = """
if ComFpy.cmdSeq.Debug.nextStatementOpcode == 0:
    # should be 0 because we aren't in debug mode
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_get_const_struct_member(fprime_test_api):
    seq = """
var: Svc.DpRecord = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED)
if var.priority == 3:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)

def test_get_const_member_of_ctor(fprime_test_api):
    seq = """
# currently this is not supported, but it should be in the future
var: U32 = Svc.DpRecord(0, 1, 2, 3, 4, 5, Fw.DpState.UNTRANSMITTED).priority
if priority == 3:
    exit(True)
exit(False)
"""

    assert_compile_failure(fprime_test_api, seq)

def test_float_cmp(fprime_test_api):
    seq = """
if 4.0 > 5.0:
    exit(False)
exit(True)
"""

    assert_run_success(fprime_test_api, seq)


def test_wait_rel(fprime_test_api):
    seq = """
sleep(1.1)
"""
    assert_run_success(fprime_test_api, seq)


def test_f32_f64_cmp(fprime_test_api):
    seq = """
val: F32 = 0.0
val2: F64 = 1.0
if val > val2:
    exit(False)
exit(True)
"""

    assert_run_success(fprime_test_api, seq)


def test_construct_array(fprime_test_api):
    seq = """
val: Svc.ComQueueDepth = Svc.ComQueueDepth(0, 0)
"""

    assert_run_success(fprime_test_api, seq)


def test_get_item_of_var(fprime_test_api):
    seq = """
val: Svc.ComQueueDepth = Svc.ComQueueDepth(0, 0)
if val[0] == 0:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_i32_f64_cmp(fprime_test_api):
    seq = """
val: I32 = 2
val2: F64 = 1.0
if val > val2:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_i32_u32_cmp(fprime_test_api):
    seq = """
val: I32 = -2
val2: U32 = 2
if val < val2:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


# caught one bug
def test_float_int_literal_cmp(fprime_test_api):
    seq = """
if 1 < 2.0:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


# caught one bug
def test_and_of_ors(fprime_test_api):
    seq = """
if True or False and True or True:
    exit(True)
exit(False)
"""

    assert_run_success(fprime_test_api, seq)


def test_if_true(fprime_test_api):
    seq = """
if True:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_if_false(fprime_test_api):
    seq = """
if False:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_if_else_true(fprime_test_api):
    seq = """
if True:
    exit(True)
else:
    exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_if_else_false(fprime_test_api):
    seq = """
if False:
    exit(False)
else:
    exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_if_elif_else(fprime_test_api):
    seq = """
if False:
    exit(False)
elif True:
    exit(True)
else:
    exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_and_true_true(fprime_test_api):
    seq = """
if True and True:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_and_true_false(fprime_test_api):
    seq = """
if True and False:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_or_false_false(fprime_test_api):
    seq = """
if False or False:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_or_true_false(fprime_test_api):
    seq = """
if True or False:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_not_true(fprime_test_api):
    seq = """
if not True:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_not_false(fprime_test_api):
    seq = """
if not False:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_complex_and_or_not(fprime_test_api):
    seq = """
if not False and (True or False):
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_literal_comparison(fprime_test_api):
    seq = """
if 255 > 254:
    exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_literal_comparison_false(fprime_test_api):
    seq = """
if 255 < 254:
    exit(False)
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_u8(fprime_test_api):
    seq = """
val1: U8 = 200
val2: U8 = 100

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_i8(fprime_test_api):
    seq = """
val1: I8 = 100
val2: I8 = -100

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_u32(fprime_test_api):
    seq = """
val1: U32 = 4294967295
val2: U32 = 0

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_i32(fprime_test_api):
    seq = """
val1: I32 = 2147483647
val2: I32 = -2147483648

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_f32(fprime_test_api):
    seq = """
val1: F32 = 3.14159
val2: F32 = -3.14159

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_all_comparison_operators_f64(fprime_test_api):
    seq = """
val1: F64 = 3.14159265359
val2: F64 = -3.14159265359

if val1 > val2 and val2 < val1:
    if val1 >= val2 and val2 <= val1:
        if val1 != val2 and not (val1 == val2):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_mixed_numeric_comparisons(fprime_test_api):
    seq = """
val_u8: U8 = 100
val_i8: I8 = -100
val_u32: U32 = 4294967295
val_i32: I32 = -2147483648
val_f32: F32 = 3.14159
val_f64: F64 = -3.14159265359

if val_u8 > val_i8 and val_i32 < val_u32:
    if val_f64 <= val_f32 and val_f32 >= val_f64:
        if val_u8 != val_i8 and not (val_u32 == val_i32):
            exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_equality_edge_cases(fprime_test_api):
    seq = """
val1: U8 = 0
val2: U8 = 0
val3: F32 = 0.0
val4: F64 = 0.0
val5: I32 = 0

if val1 == val2 and val3 == val4 and val4 == val5:
    if not (val1 != val2) and not (val3 != val4) and not (val4 != val5):
        exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_various_mixed_type_cmps(fprime_test_api):
    seq = """
val1: F32 = 3.14159
val2: I32 = -42
val3: U32 = 4294967295

if val1 > val2 and val1 < val3:  # Mixed float/signed/unsigned comparison
    if val2 < val3:  # Signed vs unsigned comparison
        exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_nested_boolean_expressions(fprime_test_api):
    seq = """
if not (True and False or True and not False) and True:
    exit(False)  # Should not execute
exit(True)
"""
    assert_run_success(fprime_test_api, seq)


def test_maximum_integer_comparisons(fprime_test_api):
    seq = """
val1: U64 = 18446744073709551615  # Max U64, interpreted as -1 in signed
val2: I64 = 9223372036854775807   # Max I64
val3: I64 = -9223372036854775808  # Min I64

if val1 < val2 and val2 > val3:
    if val3 < val1: # opposite of what you might expect, but it's cuz val1 is interpreted as signed
        exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)


def test_complex_type_assignments(fprime_test_api):
    seq = """
val1: I8 = 127
val2: U8 = 255
val3: F32 = 127.0

if val1 == val3:  # Integer to float comparison
    if val2 > val3:  # Unsigned vs float comparison
        exit(True)
exit(False)
"""
    assert_run_success(fprime_test_api, seq)



def test_type_mismatch_compile_error(fprime_test_api):
    seq = """
val1: U32 = -1  # Should fail: negative value in unsigned type
"""
    assert_compile_failure(fprime_test_api, seq)


def test_overflow_compile_error(fprime_test_api):
    seq = """
val1: U8 = 256  # Should fail: value too large for U8
"""
    assert_compile_failure(fprime_test_api, seq)


def test_mixed_boolean_numeric_comparison(fprime_test_api):
    seq = """
val1: U8 = 1
val2: I8 = -1
if (val1 > 0) == True and (val2 < 0) == True:  # Compare boolean results
    if not ((val1 <= 0) == True or (val2 >= 0) == True):
        exit(True)
exit(False)
"""
    # cannot currently compare booleans
    assert_compile_failure(fprime_test_api, seq)


def test_complex_boolean_nesting(fprime_test_api):
    seq = """
if not not not not not True:  # Multiple not operators
    exit(False)
elif not (True and not (False or not True)):  # Complex nesting
    exit(False)
else:
    exit(True)
"""
    assert_run_success(fprime_test_api, seq)


from fprime_gds.common.loaders.json_loader import JsonLoader
from fprime_gds.common.loaders.cmd_json_loader import CmdJsonLoader
from fprime_gds.common.loaders.ch_json_loader import ChJsonLoader
from fprime_gds.common.loaders.event_json_loader import EventJsonLoader
from fprime_gds.common.loaders.pkt_json_loader import PktJsonLoader
from fprime_gds.common.loaders.fw_type_json_loader import FwTypeJsonLoader
from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.enum_type import EnumType
import fprime.common.models.serialize.numerical_types as numerical_types
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.string_type import StringType

from pathlib import Path
import pytest
import json
from typing import List
from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.templates.event_template import EventTemplate
from fprime_gds.common.templates.pkt_template import PktTemplate


REF_JSON_DICTIONARY = (
    Path(__file__).resolve().parent / "resources" / "RefTopologyDictionary.json"
)

@pytest.fixture
def loader():
    return JsonLoader(REF_JSON_DICTIONARY)

@pytest.fixture
def cmd_loader():
    return CmdJsonLoader(REF_JSON_DICTIONARY)

@pytest.fixture
def event_loader():
    return EventJsonLoader(REF_JSON_DICTIONARY)

@pytest.fixture
def ch_loader():
    return ChJsonLoader(REF_JSON_DICTIONARY)

@pytest.fixture
def pkt_loader():
    return PktJsonLoader(REF_JSON_DICTIONARY)

@pytest.fixture
def fw_type_loader():
    return FwTypeJsonLoader(REF_JSON_DICTIONARY)

@pytest.fixture
def json_dict_obj():
    with open(REF_JSON_DICTIONARY, "r") as f:
        return json.load(f)


def test_construct_enum_type(loader):
    ref_signal_type = loader.parse_type(
        {"name": "Ref.SignalType", "kind": "qualifiedIdentifier"}
    )
    assert issubclass(ref_signal_type, EnumType)
    assert ref_signal_type.__name__ == "Ref.SignalType"
    assert ref_signal_type.ENUM_DICT == {
        "TRIANGLE": 0,
        "SQUARE": 1,
        "SINE": 2,
        "NOISE": 3,
    }
    assert ref_signal_type.REP_TYPE == "I32"

def test_construct_alias_type(loader):
    ref_alias_frequency_type = loader.parse_type(
        {"name" : "Ref.SignalGen.FrequencyType", "kind" : "qualifiedIdentifier"}
    )
    assert ref_alias_frequency_type == numerical_types.U32Type
    ref_alias_phase_type = loader.parse_type(
        {"name" : "Ref.SignalGen.PhaseType", "kind" : "qualifiedIdentifier"}
    )
    assert ref_alias_phase_type == numerical_types.F32Type
    ref_alias_buff_recv_type = loader.parse_type(
        {"name" : "Ref.BuffRecvType", "kind" : "qualifiedIdentifier"}
    )
    assert ref_alias_buff_recv_type == numerical_types.U32Type
    ref_alias_u32_type = loader.parse_type(
        {"name" : "Ref.AliasU32", "kind" : "qualifiedIdentifier"}
    )
    assert ref_alias_u32_type == numerical_types.U32Type

def test_construct_array_type(loader):
    ref_many_choices = loader.parse_type(
        {"name": "Ref.ManyChoices", "kind": "qualifiedIdentifier"}
    )
    assert issubclass(ref_many_choices, ArrayType)
    assert ref_many_choices.__name__ == "Ref.ManyChoices"
    assert ref_many_choices.FORMAT == "{}"
    assert ref_many_choices.LENGTH == 2
    assert ref_many_choices.MEMBER_TYPE.ENUM_DICT == {
        "ONE": 0,
        "TWO": 1,
        "RED": 2,
        "BLUE": 3,
    }
    assert ref_many_choices.MEMBER_TYPE.REP_TYPE == "I32"


def test_construct_serializable_type(loader):
    ref_choice_pair = loader.parse_type(
        {"name": "Ref.ChoicePair", "kind": "qualifiedIdentifier"}
    )
    assert issubclass(ref_choice_pair, SerializableType)
    assert ref_choice_pair.__name__ == "Ref.ChoicePair"
    assert ref_choice_pair.MEMBER_LIST[0][0] == "firstChoice"
    assert ref_choice_pair.MEMBER_LIST[0][1].ENUM_DICT == {
        "ONE": 0,
        "TWO": 1,
        "RED": 2,
        "BLUE": 3,
    }
    assert ref_choice_pair.MEMBER_LIST[0][1].REP_TYPE == "I32"
    assert ref_choice_pair.MEMBER_LIST[0][2] == "{}"
    assert ref_choice_pair.MEMBER_LIST[1][0] == "secondChoice"
    assert ref_choice_pair.MEMBER_LIST[1][1].ENUM_DICT == {
        "ONE": 0,
        "TWO": 1,
        "RED": 2,
        "BLUE": 3,
    }
    assert ref_choice_pair.MEMBER_LIST[1][1].REP_TYPE == "I32"
    assert ref_choice_pair.MEMBER_LIST[1][2] == "{}"


def test_struct_with_unordered_members(loader):
    misordered_member = loader.parse_type(
        {"name": "Ref.TestMisorderedStructIndexes", "kind": "qualifiedIdentifier"}
    )
    assert issubclass(misordered_member, SerializableType)
    assert misordered_member.MEMBER_LIST[0][0] == "ThisIsZero"
    assert misordered_member.MEMBER_LIST[1][0] == "ThisIsOne"
    assert misordered_member.MEMBER_LIST[2][0] == "ThisIsTwo"


def test_construct_primitive_types(loader):
    i32_type = loader.parse_type(
        {"name": "I32", "kind": "integer", "size": 32, "signed": True}
    )
    assert i32_type == numerical_types.I32Type
    f64_type = loader.parse_type(
        {
            "name": "F64",
            "kind": "float",
            "size": 64,
        }
    )
    assert f64_type == numerical_types.F64Type


def test_construct_cmd_dict(cmd_loader, json_dict_obj):
    id_dict, name_dict, versions = cmd_loader.construct_dicts(None)
    assert len(id_dict) == len(name_dict) == len(json_dict_obj["commands"])
    assert versions == ("TestVersion", "TestVersion")

    cmd_no_op_string: CmdTemplate = name_dict["Ref.cmdDisp.CMD_NO_OP_STRING"]
    assert cmd_no_op_string.get_op_code() == 1281
    assert cmd_no_op_string.get_description() == "No-op string command"
    assert issubclass(cmd_no_op_string.get_args()[0][2], StringType)


def test_construct_event_dict(event_loader, json_dict_obj):
    id_dict, name_dict, versions = event_loader.construct_dicts(None)
    assert len(id_dict) == len(name_dict) == len(json_dict_obj["events"])
    assert versions == ("TestVersion", "TestVersion")

    event_choice: EventTemplate = name_dict["Ref.typeDemo.ChoiceEv"]
    assert event_choice.get_id() == 4352
    assert event_choice.get_description() == "Single choice event"
    assert event_choice.get_args()[0][0] == "choice"
    assert issubclass(event_choice.get_args()[0][2], EnumType)
    assert event_choice.get_format_str() == "Choice: {}"



def test_construct_ch_dict(ch_loader, json_dict_obj):
    id_dict, name_dict, versions = ch_loader.construct_dicts(None)
    assert len(id_dict) == len(name_dict) == len(json_dict_obj["telemetryChannels"])
    assert versions == ("TestVersion", "TestVersion")

    ch_choice: ChTemplate = name_dict["Ref.typeDemo.ChoicesCh"]
    assert ch_choice.get_id() == 4353
    assert ch_choice.get_ch_desc() == "Multiple choice channel via Array"
    assert ch_choice.ch_type_obj.__name__ == "Ref.ManyChoices"
    assert ch_choice.ch_type_obj.LENGTH == 2


def test_construct_pkt_dict(ch_loader, pkt_loader):
    _, ch_name_dict, _ = ch_loader.construct_dicts(None)
    _, pkt_name_dict, _ = pkt_loader.construct_dicts("PacketSet1", ch_name_dict)
    sig_gen_pkt: PktTemplate = pkt_name_dict["SigGen1"]
    sig_gen_pkt_ch_list: List[ChTemplate] = sig_gen_pkt.get_ch_list()

    assert sig_gen_pkt.get_id() == 15
    assert len(sig_gen_pkt_ch_list) == 5
    assert sig_gen_pkt_ch_list[0].get_full_name() == "Ref.SG1.PairOutput"
    assert sig_gen_pkt_ch_list[1].get_full_name() == "Ref.SG1.History"
    assert sig_gen_pkt_ch_list[2].get_full_name() == "Ref.SG1.PairHistory"
    assert sig_gen_pkt_ch_list[3].get_full_name() == "Ref.SG1.DpBytes"
    assert sig_gen_pkt_ch_list[4].get_full_name() == "Ref.SG1.DpRecords"

    _, pkt_name_dict, _ = pkt_loader.construct_dicts("PacketSet2", ch_name_dict)
    type_demo_pkt: PktTemplate = pkt_name_dict["TypeDemo"]
    type_demo_pkt_ch_list: List[ChTemplate] = type_demo_pkt.get_ch_list()

    assert type_demo_pkt.get_id() == 21
    assert len(type_demo_pkt_ch_list) == 20
    assert type_demo_pkt_ch_list[0].get_full_name() == "Ref.typeDemo.ChoiceCh"
    assert type_demo_pkt_ch_list[1].get_full_name() == "Ref.typeDemo.ChoicesCh"
    assert type_demo_pkt_ch_list[-1].get_full_name() == "Ref.typeDemo.ScalarF64Ch"


def test_construct_fw_types(fw_type_loader):
    _, name_dict, _ = fw_type_loader.construct_dicts(None)
    assert name_dict["FwPacketDescriptorType"] == "U8" # as per specified in RefTopologyDictionary.json
    assert name_dict["FwChanIdType"] == "U32"
    assert name_dict["FwEventIdType"] == "U32"
    assert name_dict["FwOpcodeType"] == "U32"
    assert name_dict["FwTlmPacketizeIdType"] == "U16"
